"""Unit tests for FabricPipelineService."""

import base64
import json
from unittest.mock import Mock

import pytest

from ms_fabric_mcp_server.client.exceptions import (FabricAPIError,
                                                    FabricError,
                                                    FabricItemNotFoundError,
                                                    FabricValidationError)
from ms_fabric_mcp_server.models.item import FabricItem
from ms_fabric_mcp_server.services.pipeline import FabricPipelineService


def _decode_payload(payload: str) -> dict:
    return json.loads(base64.b64decode(payload).decode("utf-8"))


def _mock_response(data: dict) -> Mock:
    """Create a mock response with json() method."""
    response = Mock()
    response.json.return_value = data
    return response


@pytest.mark.unit
class TestFabricPipelineService:
    """Test suite for FabricPipelineService."""

    @pytest.fixture
    def mock_client(self):
        """Create a mock FabricClient."""
        return Mock()

    @pytest.fixture
    def mock_workspace_service(self):
        """Create a mock FabricWorkspaceService."""
        return Mock()

    @pytest.fixture
    def mock_item_service(self):
        """Create a mock FabricItemService."""
        return Mock()

    @pytest.fixture
    def pipeline_service(self, mock_client, mock_workspace_service, mock_item_service):
        """Create a FabricPipelineService instance with mocked dependencies."""
        return FabricPipelineService(
            mock_client, mock_workspace_service, mock_item_service
        )

    def test_validate_pipeline_inputs_success(self, pipeline_service):
        """Test successful validation of pipeline inputs."""
        # Should not raise any exception
        pipeline_service._validate_pipeline_inputs(
            pipeline_name="Test_Pipeline",
            source_type="AzurePostgreSqlSource",
            source_connection_id="conn-123",
            source_schema="public",
            source_table="movie",
            destination_lakehouse_id="lakehouse-456",
            destination_connection_id="dest-conn-789",
            destination_table="movie",
        )

    def test_validate_pipeline_inputs_empty_name(self, pipeline_service):
        """Test validation fails for empty pipeline name."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="lakehouse-456",
                destination_connection_id="dest-conn-789",
                destination_table="movie",
            )
        assert "pipeline_name" in str(exc_info.value)

    def test_validate_pipeline_inputs_empty_source_type(self, pipeline_service):
        """Test validation fails for empty source type."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="Test_Pipeline",
                source_type="",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="lakehouse-456",
                destination_connection_id="dest-conn-789",
                destination_table="movie",
            )
        assert "source_type" in str(exc_info.value)

    def test_validate_pipeline_inputs_empty_connection(self, pipeline_service):
        """Test validation fails for empty connection ID."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="Test_Pipeline",
                source_type="AzurePostgreSqlSource",
                source_connection_id="",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="lakehouse-456",
                destination_connection_id="dest-conn-789",
                destination_table="movie",
            )
        assert "source_connection_id" in str(exc_info.value)

    def test_validate_pipeline_inputs_empty_schema(self, pipeline_service):
        """Test validation fails for empty schema."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="Test_Pipeline",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="",
                source_table="movie",
                destination_lakehouse_id="lakehouse-456",
                destination_connection_id="dest-conn-789",
                destination_table="movie",
            )
        assert "source_schema" in str(exc_info.value)

    def test_validate_pipeline_inputs_empty_table(self, pipeline_service):
        """Test validation fails for empty source table."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="Test_Pipeline",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="",
                destination_lakehouse_id="lakehouse-456",
                destination_connection_id="dest-conn-789",
                destination_table="movie",
            )
        assert "source_table" in str(exc_info.value)

    def test_validate_pipeline_inputs_empty_destination_lakehouse(
        self, pipeline_service
    ):
        """Test validation fails for empty destination lakehouse ID."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="Test_Pipeline",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="",
                destination_connection_id="dest-conn-789",
                destination_table="movie",
            )
        assert "destination_lakehouse_id" in str(exc_info.value)

    def test_validate_pipeline_inputs_empty_destination_connection(
        self, pipeline_service
    ):
        """Test validation fails for empty destination connection ID."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="Test_Pipeline",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="lakehouse-456",
                destination_connection_id="",
                destination_table="movie",
            )
        assert "destination_connection_id" in str(exc_info.value)

    def test_validate_pipeline_inputs_empty_destination_table(self, pipeline_service):
        """Test validation fails for empty destination table."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service._validate_pipeline_inputs(
                pipeline_name="Test_Pipeline",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="lakehouse-456",
                destination_connection_id="dest-conn-789",
                destination_table="",
            )
        assert "destination_table" in str(exc_info.value)

    def test_build_copy_activity_definition(self, pipeline_service):
        """Test building Copy Activity definition."""
        definition = pipeline_service._build_copy_activity_definition(
            workspace_id="workspace-123",
            source_type="AzurePostgreSqlSource",
            source_connection_id="conn-123",
            source_schema="public",
            source_table="movie",
            destination_lakehouse_id="lakehouse-456",
            destination_connection_id="dest-conn-789",
            destination_table="movie",
            table_action_option="Append",
            apply_v_order=True,
            timeout="01:00:00",
            retry=0,
            retry_interval_seconds=30,
        )

        # Verify structure
        assert "properties" in definition
        assert "activities" in definition["properties"]
        assert len(definition["properties"]["activities"]) == 1

        # Verify activity details
        activity = definition["properties"]["activities"][0]
        assert activity["type"] == "Copy"

        # Verify source with datasetSettings
        assert activity["typeProperties"]["source"]["type"] == "AzurePostgreSqlSource"
        source_dataset = activity["typeProperties"]["source"]["datasetSettings"]
        assert source_dataset["typeProperties"]["schema"] == "public"
        assert source_dataset["typeProperties"]["table"] == "movie"
        assert source_dataset["externalReferences"]["connection"] == "conn-123"

        # Verify sink with datasetSettings
        assert activity["typeProperties"]["sink"]["type"] == "LakehouseTableSink"
        assert activity["typeProperties"]["sink"]["tableActionOption"] == "Append"
        sink_dataset = activity["typeProperties"]["sink"]["datasetSettings"]
        assert sink_dataset["type"] == "LakehouseTable"
        assert sink_dataset["typeProperties"]["table"] == "movie"

    def test_build_copy_activity_definition_lakehouse_omits_schema(
        self, pipeline_service
    ):
        """LakehouseTableSource should omit schema in source dataset settings."""
        definition = pipeline_service._build_copy_activity_definition(
            workspace_id="workspace-123",
            source_type="LakehouseTableSource",
            source_connection_id="conn-123",
            source_schema="dbo",
            source_table="fact_sale",
            destination_lakehouse_id="lakehouse-456",
            destination_connection_id="dest-conn-789",
            destination_table="fact_sale_copy",
            table_action_option="Append",
            apply_v_order=True,
            timeout="01:00:00",
            retry=0,
            retry_interval_seconds=30,
        )

        activity = definition["properties"]["activities"][0]
        source_type_properties = activity["typeProperties"]["source"][
            "datasetSettings"
        ]["typeProperties"]
        assert "schema" not in source_type_properties
        assert source_type_properties["table"] == "fact_sale"

    def test_build_copy_activity_definition_sql_mode(self, pipeline_service):
        """SQL mode should emit AzureSqlSource and AzureSqlTable with optional query."""
        definition = pipeline_service._build_copy_activity_definition(
            workspace_id="workspace-123",
            source_type="LakehouseTableSource",
            source_connection_id="conn-123",
            source_schema="dbo",
            source_table="fact_sale",
            destination_lakehouse_id="lakehouse-456",
            destination_connection_id="dest-conn-789",
            destination_table="fact_sale_copy",
            table_action_option="Append",
            apply_v_order=True,
            timeout="01:00:00",
            retry=0,
            retry_interval_seconds=30,
            source_access_mode="sql",
            source_sql_query="SELECT 1",
        )

        activity = definition["properties"]["activities"][0]
        source = activity["typeProperties"]["source"]
        assert source["type"] == "AzureSqlSource"
        assert source["datasetSettings"]["type"] == "AzureSqlTable"
        assert source["sqlReaderQuery"] == "SELECT 1"

    def test_get_source_dataset_type_mapping(self, pipeline_service):
        """Known source types map to dataset types."""
        assert (
            pipeline_service._get_source_dataset_type("AzurePostgreSqlSource")
            == "AzurePostgreSqlTable"
        )
        assert (
            pipeline_service._get_source_dataset_type("AzureSqlSource")
            == "AzureSqlTable"
        )

    def test_get_source_dataset_type_derivation(self, pipeline_service):
        """Source types ending with Source derive Table suffix."""
        assert (
            pipeline_service._get_source_dataset_type("CustomSource") == "CustomTable"
        )

    def test_get_source_dataset_type_invalid(self, pipeline_service):
        """Unsupported source types fall back to input value."""
        assert (
            pipeline_service._get_source_dataset_type("UnsupportedType")
            == "UnsupportedType"
        )

    def test_validate_source_access_mode_invalid(self, pipeline_service):
        """Invalid source access mode raises validation error."""
        with pytest.raises(FabricValidationError):
            pipeline_service._validate_source_access_mode("bad", None)

    def test_validate_source_access_mode_query_requires_sql(self, pipeline_service):
        """SQL query requires sql access mode."""
        with pytest.raises(FabricValidationError):
            pipeline_service._validate_source_access_mode("direct", "SELECT 1")

    def test_encode_definition(self, pipeline_service):
        """Test encoding pipeline definition to Base64."""
        test_definition = {"properties": {"activities": [], "parameters": {}}}

        encoded = pipeline_service._encode_definition(test_definition)

        # Verify it's a valid base64 string
        assert isinstance(encoded, str)

        # Verify we can decode it back
        decoded_bytes = base64.b64decode(encoded)
        decoded_str = decoded_bytes.decode("utf-8")
        decoded_obj = json.loads(decoded_str)

        assert decoded_obj == test_definition

    def test_decode_definition_round_trip(self, pipeline_service):
        """Encode/decode round trip returns original definition."""
        test_definition = {"properties": {"activities": [{"name": "A1"}]}}

        encoded = pipeline_service._encode_definition(test_definition)
        decoded = pipeline_service._decode_definition(encoded)

        assert decoded == test_definition

    def test_decode_definition_invalid_payload(self, pipeline_service):
        """Invalid payload raises FabricError."""
        with pytest.raises(FabricError):
            pipeline_service._decode_definition("not-base64")

    def test_create_blank_pipeline_success(self, pipeline_service, mock_item_service):
        """Create blank pipeline builds definition and returns ID."""
        mock_item_service.create_item.return_value = FabricItem(
            id="pipe-1",
            display_name="BlankPipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )

        pipeline_id = pipeline_service.create_blank_pipeline(
            workspace_id="ws-1",
            pipeline_name="BlankPipe",
            description="desc",
        )

        assert pipeline_id == "pipe-1"
        args, kwargs = mock_item_service.create_item.call_args
        item_definition = args[1]
        assert item_definition["displayName"] == "BlankPipe"
        assert item_definition["type"] == "DataPipeline"
        assert item_definition["description"] == "desc"
        payload = item_definition["definition"]["parts"][0]["payload"]
        decoded = _decode_payload(payload)
        assert decoded["properties"]["activities"] == []

    def test_create_blank_pipeline_with_folder_path(
        self, pipeline_service, mock_item_service
    ):
        """Folder path resolves to folderId on create."""
        mock_item_service.create_item.return_value = FabricItem(
            id="pipe-2",
            display_name="BlankPipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        mock_item_service.resolve_folder_id_from_path.return_value = "folder-1"

        pipeline_service.create_blank_pipeline(
            workspace_id="ws-1",
            pipeline_name="BlankPipe",
            folder_path="Pipelines/Daily",
        )

        mock_item_service.resolve_folder_id_from_path.assert_called_once_with(
            "ws-1", "Pipelines/Daily", create_missing=True
        )
        item_definition = mock_item_service.create_item.call_args.args[1]
        assert item_definition["folderId"] == "folder-1"

    def test_create_blank_pipeline_validation_error(self, pipeline_service):
        """Empty pipeline name raises validation error."""
        with pytest.raises(FabricValidationError):
            pipeline_service.create_blank_pipeline("ws-1", " ")

    def test_create_blank_pipeline_invalid_name(self, pipeline_service):
        """Names with separators raise validation error."""
        with pytest.raises(FabricValidationError):
            pipeline_service.create_blank_pipeline("ws-1", "Bad/Name")

    def test_create_blank_pipeline_api_error(self, pipeline_service, mock_item_service):
        """API errors propagate."""
        mock_item_service.create_item.side_effect = FabricAPIError(500, "boom")

        with pytest.raises(FabricAPIError):
            pipeline_service.create_blank_pipeline("ws-1", "Pipe")

    def test_create_blank_pipeline_unexpected_error(
        self, pipeline_service, mock_item_service
    ):
        """Unexpected errors raise FabricError."""
        mock_item_service.create_item.side_effect = RuntimeError("boom")

        with pytest.raises(FabricError):
            pipeline_service.create_blank_pipeline("ws-1", "Pipe")

    def test_create_pipeline_with_definition_success(
        self, pipeline_service, mock_client
    ):
        """Create pipeline with definition posts encoded parts and returns ID."""
        response = Mock()
        response.status_code = 201
        response.json.return_value = {"id": "pipe-123", "displayName": "Pipe"}
        mock_client.make_api_request.return_value = response

        pipeline_id = pipeline_service.create_pipeline_with_definition(
            workspace_id="ws-1",
            display_name="Pipe",
            pipeline_content_json={"properties": {"activities": []}},
            platform={"metadata": {"key": "value"}},
            description="desc",
            folder_id="folder-1",
        )

        assert pipeline_id == "pipe-123"
        args, kwargs = mock_client.make_api_request.call_args
        assert args[0] == "POST"
        assert args[1] == "workspaces/ws-1/dataPipelines"
        payload = kwargs["payload"]
        assert payload["displayName"] == "Pipe"
        assert payload["description"] == "desc"
        assert payload["folderId"] == "folder-1"
        parts = payload["definition"]["parts"]
        assert parts[0]["path"] == "pipeline-content.json"
        assert _decode_payload(parts[0]["payload"]) == {
            "properties": {"activities": []}
        }
        assert parts[1]["path"] == ".platform"
        assert _decode_payload(parts[1]["payload"]) == {"metadata": {"key": "value"}}

    def test_create_pipeline_with_definition_invalid_definition(self, pipeline_service):
        """Non-dict definition raises validation error."""
        with pytest.raises(FabricValidationError):
            pipeline_service.create_pipeline_with_definition(
                workspace_id="ws-1",
                display_name="Pipe",
                pipeline_content_json="not-a-dict",
            )

    def test_create_pipeline_with_definition_invalid_name(self, pipeline_service):
        """Names with separators raise validation error."""
        with pytest.raises(FabricValidationError):
            pipeline_service.create_pipeline_with_definition(
                workspace_id="ws-1",
                display_name="Bad/Name",
                pipeline_content_json={"properties": {"activities": []}},
            )

    def test_create_pipeline_with_definition_invalid_platform(self, pipeline_service):
        """Non-dict platform raises validation error."""
        with pytest.raises(FabricValidationError):
            pipeline_service.create_pipeline_with_definition(
                workspace_id="ws-1",
                display_name="Pipe",
                pipeline_content_json={"properties": {"activities": []}},
                platform="not-a-dict",
            )

    def test_get_pipeline_definition_success(self, pipeline_service, mock_client):
        """Fetches and decodes pipeline definition with optional platform."""
        definition_payload = {"properties": {"activities": []}}
        encoded_definition = pipeline_service._encode_definition(definition_payload)
        platform_payload = {"metadata": {"key": "value"}}
        encoded_platform = pipeline_service._encode_definition(platform_payload)

        response = Mock()
        response.json.return_value = {
            "definition": {
                "parts": [
                    {"path": "pipeline-content.json", "payload": encoded_definition},
                    {"path": ".platform", "payload": encoded_platform},
                ]
            }
        }
        mock_client.make_api_request.return_value = response

        result = pipeline_service.get_pipeline_definition(
            workspace_id="ws-1",
            pipeline_id="pipe-1",
            format="json",
        )

        assert result["pipeline_content_json"] == definition_payload
        assert result["platform"] == platform_payload

        args, kwargs = mock_client.make_api_request.call_args
        assert args[0] == "POST"
        assert "workspaces/ws-1/dataPipelines/pipe-1/getDefinition" in args[1]
        assert kwargs["wait_for_lro"] is True

    def test_get_pipeline_definition_missing_content(
        self, pipeline_service, mock_client
    ):
        """Missing pipeline-content.json raises FabricError."""
        response = Mock()
        response.json.return_value = {"definition": {"parts": []}}
        mock_client.make_api_request.return_value = response

        with pytest.raises(FabricError):
            pipeline_service.get_pipeline_definition("ws-1", "pipe-1")

    def test_get_pipeline_definition_platform_fallback(
        self, pipeline_service, mock_client
    ):
        """Invalid platform JSON returns raw string."""
        definition_payload = {"properties": {"activities": []}}
        encoded_definition = pipeline_service._encode_definition(definition_payload)
        encoded_platform = base64.b64encode(b"not-json").decode("utf-8")

        response = Mock()
        response.json.return_value = {
            "definition": {
                "parts": [
                    {"path": "pipeline-content.json", "payload": encoded_definition},
                    {"path": ".platform", "payload": encoded_platform},
                ]
            }
        }
        mock_client.make_api_request.return_value = response

        result = pipeline_service.get_pipeline_definition("ws-1", "pipe-1")
        assert result["platform"] == "not-json"

    def test_update_pipeline_definition_success(self, pipeline_service, mock_client):
        """Update pipeline definition builds correct payload."""
        pipeline_content = {"properties": {"activities": []}}
        platform = {"metadata": {"key": "value"}}

        pipeline_service.update_pipeline_definition(
            workspace_id="ws-1",
            pipeline_id="pipe-1",
            pipeline_content_json=pipeline_content,
            platform=platform,
            update_metadata=True,
        )

        args, kwargs = mock_client.make_api_request.call_args
        assert args[0] == "POST"
        assert "workspaces/ws-1/dataPipelines/pipe-1/updateDefinition" in args[1]
        assert "updateMetadata=true" in args[1]
        payload = kwargs["payload"]["definition"]["parts"]
        assert payload[0]["path"] == "pipeline-content.json"
        assert _decode_payload(payload[0]["payload"]) == pipeline_content
        assert payload[1]["path"] == ".platform"
        assert _decode_payload(payload[1]["payload"]) == platform

    def test_update_pipeline_definition_requires_platform_for_metadata(
        self, pipeline_service
    ):
        """update_metadata requires platform payload."""
        with pytest.raises(FabricValidationError):
            pipeline_service.update_pipeline_definition(
                workspace_id="ws-1",
                pipeline_id="pipe-1",
                pipeline_content_json={"properties": {"activities": []}},
                update_metadata=True,
            )

    def test_set_activity_dependency_add(self, pipeline_service, mock_item_service):
        """Add dependency to activity."""
        pipeline_item = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_by_name.return_value = pipeline_item

        definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {"name": "B", "dependsOn": []},
                ]
            }
        }
        pipeline_service.get_pipeline_definition = Mock(
            return_value={"pipeline_content_json": definition}
        )
        pipeline_service.update_pipeline_definition = Mock()

        pipeline_id, changed = pipeline_service.set_activity_dependency(
            workspace_id="ws-1",
            pipeline_name="Pipe",
            activity_name="B",
            depends_on=["A"],
            mode="add",
        )

        assert pipeline_id == "pipe-1"
        assert changed == 1
        update_kwargs = pipeline_service.update_pipeline_definition.call_args.kwargs
        updated = update_kwargs["pipeline_content_json"]
        activity_b = next(
            activity
            for activity in updated["properties"]["activities"]
            if activity["name"] == "B"
        )
        assert activity_b["dependsOn"][0]["activity"] == "A"

    def test_set_activity_dependency_remove(self, pipeline_service, mock_item_service):
        """Remove dependency from activity."""
        pipeline_item = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_by_name.return_value = pipeline_item

        definition = {
            "properties": {
                "activities": [
                    {
                        "name": "A",
                        "dependsOn": [
                            {"activity": "B", "dependencyConditions": ["Succeeded"]}
                        ],
                    },
                    {"name": "B", "dependsOn": []},
                ]
            }
        }
        pipeline_service.get_pipeline_definition = Mock(
            return_value={"pipeline_content_json": definition}
        )
        pipeline_service.update_pipeline_definition = Mock()

        _, changed = pipeline_service.set_activity_dependency(
            workspace_id="ws-1",
            pipeline_name="Pipe",
            activity_name="A",
            depends_on=["B"],
            mode="remove",
        )

        assert changed == 1
        updated = pipeline_service.update_pipeline_definition.call_args.kwargs[
            "pipeline_content_json"
        ]
        activity_a = next(
            activity
            for activity in updated["properties"]["activities"]
            if activity["name"] == "A"
        )
        assert activity_a["dependsOn"] == []

    def test_set_activity_dependency_replace(self, pipeline_service, mock_item_service):
        """Replace dependencies for activity."""
        pipeline_item = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_by_name.return_value = pipeline_item

        definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": [{"activity": "B"}]},
                    {"name": "B", "dependsOn": []},
                    {"name": "C", "dependsOn": []},
                ]
            }
        }
        pipeline_service.get_pipeline_definition = Mock(
            return_value={"pipeline_content_json": definition}
        )
        pipeline_service.update_pipeline_definition = Mock()

        _, changed = pipeline_service.set_activity_dependency(
            workspace_id="ws-1",
            pipeline_name="Pipe",
            activity_name="A",
            depends_on=["C"],
            mode="replace",
            dependency_conditions=["Succeeded", "Failed"],
        )

        assert changed == 1
        updated = pipeline_service.update_pipeline_definition.call_args.kwargs[
            "pipeline_content_json"
        ]
        activity_a = next(
            activity
            for activity in updated["properties"]["activities"]
            if activity["name"] == "A"
        )
        assert activity_a["dependsOn"][0]["activity"] == "C"
        assert activity_a["dependsOn"][0]["dependencyConditions"] == [
            "Succeeded",
            "Failed",
        ]

    def test_add_copy_activity_to_pipeline_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Adds copy activity and updates definition."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {"properties": {"activities": []}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id = pipeline_service.add_copy_activity_to_pipeline(
            workspace_id="ws-1",
            pipeline_name="Pipe",
            source_type="AzurePostgreSqlSource",
            source_connection_id="conn-123",
            source_schema="public",
            source_table="movie",
            destination_lakehouse_id="lh-1",
            destination_connection_id="dest-conn",
            destination_table="movie",
            activity_name="CopyMovieData",
        )

        assert pipeline_id == "pipe-1"
        _, kwargs = mock_client.make_api_request.call_args
        update_payload = kwargs["payload"]
        payload = update_payload["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        assert updated["properties"]["activities"][-1]["name"] == "CopyMovieData"

    def test_add_copy_activity_to_pipeline_missing_part(
        self, pipeline_service, mock_item_service
    ):
        """Missing pipeline-content.json raises FabricError."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_definition.return_value = {
            "definition": {"parts": []}
        }

        with pytest.raises(FabricError):
            pipeline_service.add_copy_activity_to_pipeline(
                workspace_id="ws-1",
                pipeline_name="Pipe",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="lh-1",
                destination_connection_id="dest-conn",
                destination_table="movie",
            )

    def test_add_copy_activity_to_pipeline_api_error(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """API errors propagate."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {"properties": {"activities": []}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }
        mock_client.make_api_request.side_effect = FabricAPIError(500, "boom")

        with pytest.raises(FabricAPIError):
            pipeline_service.add_copy_activity_to_pipeline(
                workspace_id="ws-1",
                pipeline_name="Pipe",
                source_type="AzurePostgreSqlSource",
                source_connection_id="conn-123",
                source_schema="public",
                source_table="movie",
                destination_lakehouse_id="lh-1",
                destination_connection_id="dest-conn",
                destination_table="movie",
            )

    def test_add_notebook_activity_to_pipeline_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Adds notebook activity and updates definition."""
        pipeline_item = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        notebook_item = FabricItem(
            id="nb-1",
            display_name="Note",
            type="Notebook",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_by_name.side_effect = [pipeline_item, notebook_item]
        base_definition = {"properties": {"activities": []}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id = pipeline_service.add_notebook_activity_to_pipeline(
            workspace_id="ws-1",
            pipeline_name="Pipe",
            notebook_name="Note",
            activity_name="RunNotebook_Note",
            session_tag="tag-1",
            parameters={"p1": {"value": "x", "type": "string"}},
        )

        assert pipeline_id == "pipe-1"
        _, kwargs = mock_client.make_api_request.call_args
        payload = kwargs["payload"]["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        activity = updated["properties"]["activities"][-1]
        assert activity["type"] == "TridentNotebook"
        assert activity["typeProperties"]["notebookId"] == "nb-1"
        assert activity["typeProperties"]["workspaceId"] == "ws-1"
        assert activity["typeProperties"]["sessionTag"] == "tag-1"
        assert activity["typeProperties"]["parameters"]["p1"]["value"] == "x"

    def test_add_notebook_activity_to_pipeline_dependency_missing(
        self, pipeline_service, mock_item_service
    ):
        """Missing dependency raises validation error."""
        pipeline_item = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        notebook_item = FabricItem(
            id="nb-1",
            display_name="Note",
            type="Notebook",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_by_name.side_effect = [pipeline_item, notebook_item]
        base_definition = {"properties": {"activities": [{"name": "Existing"}]}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.add_notebook_activity_to_pipeline(
                workspace_id="ws-1",
                pipeline_name="Pipe",
                notebook_name="Note",
                activity_name="RunNotebook_Note",
                depends_on_activity_name="MissingActivity",
            )

    def test_add_dataflow_activity_to_pipeline_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Adds dataflow activity and updates definition."""
        pipeline_item = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        dataflow_item = FabricItem(
            id="df-1",
            display_name="Flow",
            type="Dataflow",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_by_name.side_effect = [pipeline_item, dataflow_item]
        base_definition = {"properties": {"activities": []}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id = pipeline_service.add_dataflow_activity_to_pipeline(
            workspace_id="ws-1",
            pipeline_name="Pipe",
            dataflow_name="Flow",
            activity_name="RunDataflow_Flow",
        )

        assert pipeline_id == "pipe-1"
        _, kwargs = mock_client.make_api_request.call_args
        payload = kwargs["payload"]["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        activity = updated["properties"]["activities"][-1]
        assert activity["type"] == "RefreshDataflow"
        assert activity["typeProperties"]["dataflowId"] == "df-1"
        assert activity["typeProperties"]["workspaceId"] == "ws-1"
        assert activity["typeProperties"]["dataflowType"] == "Dataflow-Gen2"

    def test_add_dataflow_activity_to_pipeline_duplicate_name(
        self, pipeline_service, mock_item_service
    ):
        """Duplicate activity names raise validation error."""
        pipeline_item = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        dataflow_item = FabricItem(
            id="df-1",
            display_name="Flow",
            type="Dataflow",
            workspace_id="ws-1",
        )
        mock_item_service.get_item_by_name.side_effect = [pipeline_item, dataflow_item]
        base_definition = {"properties": {"activities": [{"name": "RunDataflow_Flow"}]}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.add_dataflow_activity_to_pipeline(
                workspace_id="ws-1",
                pipeline_name="Pipe",
                dataflow_name="Flow",
                activity_name="RunDataflow_Flow",
            )

    def test_add_activity_from_json_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Adds generic activity and updates definition."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {"properties": {"activities": []}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }
        activity = {"name": "MyActivity", "type": "Custom", "dependsOn": []}

        pipeline_id = pipeline_service.add_activity_from_json("ws-1", "Pipe", activity)

        assert pipeline_id == "pipe-1"
        _, kwargs = mock_client.make_api_request.call_args
        update_payload = kwargs["payload"]
        payload = update_payload["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        assert updated["properties"]["activities"][-1]["name"] == "MyActivity"

    def test_add_activity_from_json_validation(self, pipeline_service):
        """Missing name/type in activity_json raises validation error."""
        with pytest.raises(FabricValidationError):
            pipeline_service.add_activity_from_json("ws-1", "Pipe", "not-a-dict")
        with pytest.raises(FabricValidationError):
            pipeline_service.add_activity_from_json("ws-1", "Pipe", {"type": "Copy"})
        with pytest.raises(FabricValidationError):
            pipeline_service.add_activity_from_json("ws-1", "Pipe", {"name": "A1"})

    def test_add_activity_from_json_item_not_found(
        self, pipeline_service, mock_item_service
    ):
        """Item not found errors propagate."""
        mock_item_service.get_item_by_name.side_effect = FabricItemNotFoundError(
            "DataPipeline",
            "Pipe",
            "ws-1",
        )

        with pytest.raises(FabricItemNotFoundError):
            pipeline_service.add_activity_from_json(
                "ws-1", "Pipe", {"name": "A1", "type": "Copy"}
            )

    def test_delete_activity_from_pipeline_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Deletes an activity and updates definition."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {"name": "B", "dependsOn": []},
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id = pipeline_service.delete_activity_from_pipeline(
            "ws-1", "Pipe", "A"
        )

        assert pipeline_id == "pipe-1"
        _, kwargs = mock_client.make_api_request.call_args
        payload = kwargs["payload"]["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        names = [
            activity.get("name") for activity in updated["properties"]["activities"]
        ]
        assert "A" not in names
        assert "B" in names

    def test_delete_activity_from_pipeline_missing_activity(
        self, pipeline_service, mock_item_service
    ):
        """Missing activity name raises validation error."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {"activities": [{"name": "B", "dependsOn": []}]}
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.delete_activity_from_pipeline("ws-1", "Pipe", "A")

    def test_delete_activity_from_pipeline_with_dependents(
        self, pipeline_service, mock_item_service
    ):
        """Dependent activities block deletion."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {
                        "name": "B",
                        "dependsOn": [
                            {"activity": "A", "dependencyConditions": ["Succeeded"]}
                        ],
                    },
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.delete_activity_from_pipeline("ws-1", "Pipe", "A")

    def test_delete_activity_from_pipeline_duplicate_names(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Deletes all activities with a matching name."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {"name": "A", "dependsOn": []},
                    {"name": "B", "dependsOn": []},
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id = pipeline_service.delete_activity_from_pipeline(
            "ws-1", "Pipe", "A"
        )

        assert pipeline_id == "pipe-1"
        _, kwargs = mock_client.make_api_request.call_args
        payload = kwargs["payload"]["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        names = [
            activity.get("name") for activity in updated["properties"]["activities"]
        ]
        assert names == ["B"]

    def test_remove_activity_dependency_global_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Removes dependency edges across all activities."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {
                        "name": "B",
                        "dependsOn": [
                            {"activity": "A", "dependencyConditions": ["Succeeded"]},
                            {"activity": "C", "dependencyConditions": ["Succeeded"]},
                        ],
                    },
                    {
                        "name": "C",
                        "dependsOn": [
                            {"activity": "A", "dependencyConditions": ["Succeeded"]}
                        ],
                    },
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id, removed_count = pipeline_service.remove_activity_dependency(
            "ws-1", "Pipe", "A"
        )

        assert pipeline_id == "pipe-1"
        assert removed_count == 2
        _, kwargs = mock_client.make_api_request.call_args
        payload = kwargs["payload"]["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        activities = {
            activity["name"]: activity
            for activity in updated["properties"]["activities"]
        }
        assert activities["B"]["dependsOn"] == [
            {"activity": "C", "dependencyConditions": ["Succeeded"]}
        ]
        assert activities["C"]["dependsOn"] == []

    def test_remove_activity_dependency_specific_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Removes dependency edges from a specific activity."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {
                        "name": "B",
                        "dependsOn": [
                            {"activity": "A", "dependencyConditions": ["Succeeded"]}
                        ],
                    },
                    {
                        "name": "C",
                        "dependsOn": [
                            {"activity": "A", "dependencyConditions": ["Succeeded"]}
                        ],
                    },
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id, removed_count = pipeline_service.remove_activity_dependency(
            "ws-1", "Pipe", "A", from_activity_name="B"
        )

        assert pipeline_id == "pipe-1"
        assert removed_count == 1
        _, kwargs = mock_client.make_api_request.call_args
        payload = kwargs["payload"]["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        activities = {
            activity["name"]: activity
            for activity in updated["properties"]["activities"]
        }
        assert activities["B"]["dependsOn"] == []
        assert activities["C"]["dependsOn"] == [
            {"activity": "A", "dependencyConditions": ["Succeeded"]}
        ]

    def test_remove_activity_dependency_no_edges(
        self, pipeline_service, mock_item_service
    ):
        """No dependency edges removed raises validation error."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {
                        "name": "B",
                        "dependsOn": [
                            {"activity": "C", "dependencyConditions": ["Succeeded"]}
                        ],
                    },
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.remove_activity_dependency("ws-1", "Pipe", "A")

    def test_remove_activity_dependency_missing_target_succeeds(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Removes edges even if target activity is missing from activities list."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "B", "dependsOn": [{"activity": "Missing"}]},
                    {"name": "C", "dependsOn": []},
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id, removed_count = pipeline_service.remove_activity_dependency(
            "ws-1", "Pipe", "Missing"
        )

        assert pipeline_id == "pipe-1"
        assert removed_count == 1
        _, kwargs = mock_client.make_api_request.call_args
        payload = kwargs["payload"]["definition"]["parts"][0]["payload"]
        updated = _decode_payload(payload)
        activities = {
            activity["name"]: activity
            for activity in updated["properties"]["activities"]
        }
        assert activities["B"]["dependsOn"] == []

    def test_add_activity_dependency_success(
        self, pipeline_service, mock_item_service, mock_client
    ):
        """Adds dependsOn entries to an activity."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {"name": "A", "dependsOn": []},
                    {"name": "B", "dependsOn": []},
                    {"name": "C", "dependsOn": []},
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        pipeline_id, added_count = pipeline_service.add_activity_dependency(
            workspace_id="ws-1",
            pipeline_name="Pipe",
            activity_name="C",
            depends_on=["A", "B"],
        )

        assert pipeline_id == "pipe-1"
        assert added_count == 2
        payload = mock_client.make_api_request.call_args.kwargs["payload"]
        updated = _decode_payload(payload["definition"]["parts"][0]["payload"])
        depends_on = updated["properties"]["activities"][2]["dependsOn"]
        assert {"activity": "A", "dependencyConditions": ["Succeeded"]} in depends_on
        assert {"activity": "B", "dependencyConditions": ["Succeeded"]} in depends_on

    def test_add_activity_dependency_missing_target(
        self, pipeline_service, mock_item_service
    ):
        """Missing target activity raises validation error."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {"properties": {"activities": [{"name": "A"}]}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.add_activity_dependency(
                workspace_id="ws-1",
                pipeline_name="Pipe",
                activity_name="C",
                depends_on=["A"],
            )

    def test_add_activity_dependency_missing_dependency(
        self, pipeline_service, mock_item_service
    ):
        """Missing dependency activity raises validation error."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {"properties": {"activities": [{"name": "A"}, {"name": "C"}]}}
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.add_activity_dependency(
                workspace_id="ws-1",
                pipeline_name="Pipe",
                activity_name="C",
                depends_on=["Missing"],
            )

    def test_add_activity_dependency_no_new(self, pipeline_service, mock_item_service):
        """No new dependencies raises validation error."""
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="pipe-1",
            display_name="Pipe",
            type="DataPipeline",
            workspace_id="ws-1",
        )
        base_definition = {
            "properties": {
                "activities": [
                    {
                        "name": "C",
                        "dependsOn": [
                            {"activity": "A", "dependencyConditions": ["Succeeded"]}
                        ],
                    },
                    {"name": "A"},
                ]
            }
        }
        encoded = pipeline_service._encode_definition(base_definition)
        mock_item_service.get_item_definition.return_value = {
            "definition": {
                "parts": [{"path": "pipeline-content.json", "payload": encoded}]
            }
        }

        with pytest.raises(FabricValidationError):
            pipeline_service.add_activity_dependency(
                workspace_id="ws-1",
                pipeline_name="Pipe",
                activity_name="C",
                depends_on=["A"],
            )


@pytest.mark.unit
class TestGetPipelineActivityRuns:
    """Tests for get_pipeline_activity_runs method."""

    @pytest.fixture
    def mock_client(self):
        """Create a mock FabricClient."""
        return Mock()

    @pytest.fixture
    def mock_workspace_service(self):
        """Create a mock FabricWorkspaceService."""
        return Mock()

    @pytest.fixture
    def mock_item_service(self):
        """Create a mock FabricItemService."""
        return Mock()

    @pytest.fixture
    def pipeline_service(self, mock_client, mock_workspace_service, mock_item_service):
        """Create a FabricPipelineService instance with mocked dependencies."""
        return FabricPipelineService(
            mock_client, mock_workspace_service, mock_item_service
        )

    def test_get_pipeline_activity_runs_success(self, pipeline_service, mock_client):
        """Test successful retrieval of activity runs."""
        mock_client.make_api_request.return_value = _mock_response(
            {
                "value": [
                    {
                        "activityName": "CopyProducts",
                        "activityType": "Copy",
                        "status": "Succeeded",
                        "durationInMs": 5230,
                        "activityRunStart": "2026-02-03T13:00:00Z",
                        "activityRunEnd": "2026-02-03T13:00:05Z",
                        "pipelineName": "Copy_Data_Pipeline",
                        "output": {"rowsRead": 1234, "rowsCopied": 1234},
                    },
                    {
                        "activityName": "TransformData",
                        "activityType": "Notebook",
                        "status": "Succeeded",
                        "durationInMs": 15000,
                        "activityRunStart": "2026-02-03T13:00:05Z",
                        "activityRunEnd": "2026-02-03T13:00:20Z",
                        "pipelineName": "Copy_Data_Pipeline",
                    },
                ]
            }
        )

        result = pipeline_service.get_pipeline_activity_runs(
            workspace_id="ws-1",
            job_instance_id="job-123",
        )

        assert result["activity_count"] == 2
        assert result["pipeline_name"] == "Copy_Data_Pipeline"
        assert len(result["activities"]) == 2

        # Check Copy activity
        copy_activity = result["activities"][0]
        assert copy_activity["activity_name"] == "CopyProducts"
        assert copy_activity["activity_type"] == "Copy"
        assert copy_activity["status"] == "Succeeded"
        assert copy_activity["duration_ms"] == 5230
        assert copy_activity["rows_read"] == 1234
        assert copy_activity["rows_written"] == 1234
        assert copy_activity["error_message"] is None

        # Check Notebook activity (no row counts)
        notebook_activity = result["activities"][1]
        assert notebook_activity["activity_name"] == "TransformData"
        assert notebook_activity["rows_read"] is None
        assert notebook_activity["rows_written"] is None

    def test_get_pipeline_activity_runs_with_status_filter(
        self, pipeline_service, mock_client
    ):
        """Test retrieval with status filter."""
        mock_client.make_api_request.return_value = _mock_response({"value": []})

        pipeline_service.get_pipeline_activity_runs(
            workspace_id="ws-1",
            job_instance_id="job-123",
            status_filter="Failed",
        )

        # Verify filter was passed
        call_args = mock_client.make_api_request.call_args
        payload = call_args[1]["payload"]
        assert "filters" in payload
        assert payload["filters"][0]["operand"] == "Status"
        assert payload["filters"][0]["values"] == ["Failed"]

    def test_get_pipeline_activity_runs_with_activity_name_filter(
        self, pipeline_service, mock_client
    ):
        """Test retrieval with activity name filter."""
        mock_client.make_api_request.return_value = _mock_response({"value": []})

        pipeline_service.get_pipeline_activity_runs(
            workspace_id="ws-1",
            job_instance_id="job-123",
            activity_name_filter="CopyProducts",
        )

        # Verify filter was passed
        call_args = mock_client.make_api_request.call_args
        payload = call_args[1]["payload"]
        assert "filters" in payload
        assert payload["filters"][0]["operand"] == "ActivityName"
        assert payload["filters"][0]["values"] == ["CopyProducts"]

    def test_get_pipeline_activity_runs_failed_activity(
        self, pipeline_service, mock_client
    ):
        """Test retrieval of failed activity with error message."""
        mock_client.make_api_request.return_value = _mock_response(
            {
                "value": [
                    {
                        "activityName": "CopyProducts",
                        "activityType": "Copy",
                        "status": "Failed",
                        "durationInMs": 1230,
                        "activityRunStart": "2026-02-03T13:00:00Z",
                        "activityRunEnd": "2026-02-03T13:00:01Z",
                        "pipelineName": "Copy_Data_Pipeline",
                        "error": {"message": "Connection timeout", "code": "TIMEOUT"},
                    },
                ]
            }
        )

        result = pipeline_service.get_pipeline_activity_runs(
            workspace_id="ws-1",
            job_instance_id="job-123",
        )

        activity = result["activities"][0]
        assert activity["status"] == "Failed"
        assert activity["error_message"] == "Connection timeout"

    def test_get_pipeline_activity_runs_empty_workspace_id(self, pipeline_service):
        """Test validation error for empty workspace ID."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service.get_pipeline_activity_runs(
                workspace_id="",
                job_instance_id="job-123",
            )
        assert "workspace_id" in str(exc_info.value)

    def test_get_pipeline_activity_runs_empty_job_instance_id(self, pipeline_service):
        """Test validation error for empty job instance ID."""
        with pytest.raises(FabricValidationError) as exc_info:
            pipeline_service.get_pipeline_activity_runs(
                workspace_id="ws-1",
                job_instance_id="",
            )
        assert "job_instance_id" in str(exc_info.value)

    def test_get_pipeline_activity_runs_api_error(self, pipeline_service, mock_client):
        """Test API error propagation."""
        mock_client.make_api_request.side_effect = FabricAPIError(
            status_code=500, message="API Error"
        )

        with pytest.raises(FabricAPIError):
            pipeline_service.get_pipeline_activity_runs(
                workspace_id="ws-1",
                job_instance_id="job-123",
            )



"""Integration tests for pipeline tools."""

import asyncio
import json

import pytest

from tests.conftest import unique_name


@pytest.mark.integration
@pytest.mark.asyncio
async def test_create_pipeline_and_add_activity(
    call_tool, delete_item_if_exists, workspace_name
):
    pipeline_name = unique_name("e2e_pipeline")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline",
        )
        assert create_result["status"] == "success"

        activity = {
            "name": "WaitShort",
            "type": "Wait",
            "dependsOn": [],
            "typeProperties": {"waitTimeInSeconds": 1},
        }
        add_result = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity,
        )
        assert add_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_pipeline_definition_roundtrip(
    call_tool,
    delete_item_if_exists,
    workspace_name,
):
    pipeline_name = unique_name("e2e_pipeline_definition")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            pipeline_content_json={"properties": {"activities": []}},
        )
        assert create_result["status"] == "success"

        get_result = await call_tool(
            "get_pipeline_definition",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
        )
        assert get_result["status"] == "success"
        definition = get_result.get("pipeline_content_json", {})
        activities = definition.get("properties", {}).get("activities", [])
        assert isinstance(activities, list)

        updated_definition = {
            **definition,
            "properties": {
                **definition.get("properties", {}),
                "activities": activities
                + [
                    {
                        "name": "WaitShort",
                        "type": "Wait",
                        "dependsOn": [],
                        "typeProperties": {"waitTimeInSeconds": 1},
                    }
                ],
            },
        }

        update_result = await call_tool(
            "update_pipeline_definition",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            pipeline_content_json=updated_definition,
        )
        assert update_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_add_copy_activity_to_pipeline(
    call_tool,
    delete_item_if_exists,
    lakehouse_id,
    pipeline_copy_inputs,
    workspace_name,
):
    if not pipeline_copy_inputs:
        pytest.skip("Missing pipeline copy inputs")

    pipeline_name = unique_name("e2e_pipeline_copy")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline copy",
        )
        assert create_result["status"] == "success"

        add_result = await call_tool(
            "add_copy_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            source_type=pipeline_copy_inputs["source_type"],
            source_connection_id=pipeline_copy_inputs["source_connection_id"],
            source_table_schema=pipeline_copy_inputs["source_schema"],
            source_table_name=pipeline_copy_inputs["source_table"],
            destination_lakehouse_id=lakehouse_id,
            destination_connection_id=pipeline_copy_inputs["destination_connection_id"],
            destination_table_name=pipeline_copy_inputs["destination_table"],
        )
        assert add_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_add_copy_activity_to_pipeline_sql_mode(
    call_tool,
    delete_item_if_exists,
    lakehouse_id,
    pipeline_copy_sql_inputs,
    workspace_name,
):
    if not pipeline_copy_sql_inputs:
        pytest.skip("Missing pipeline SQL copy inputs")

    pipeline_name = unique_name("e2e_pipeline_copy_sql")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline copy (sql mode)",
        )
        assert create_result["status"] == "success"

        add_kwargs = {
            "workspace_name": workspace_name,
            "pipeline_name": pipeline_name,
            "source_type": "LakehouseTableSource",
            "source_connection_id": pipeline_copy_sql_inputs["source_connection_id"],
            "source_table_schema": pipeline_copy_sql_inputs["source_schema"],
            "source_table_name": pipeline_copy_sql_inputs["source_table"],
            "destination_lakehouse_id": lakehouse_id,
            "destination_connection_id": pipeline_copy_sql_inputs[
                "destination_connection_id"
            ],
            "destination_table_name": pipeline_copy_sql_inputs["destination_table"],
            "source_access_mode": "sql",
        }
        if pipeline_copy_sql_inputs.get("source_sql_query"):
            add_kwargs["source_sql_query"] = pipeline_copy_sql_inputs[
                "source_sql_query"
            ]

        add_result = await call_tool("add_copy_activity_to_pipeline", **add_kwargs)
        assert add_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_add_notebook_activity_to_pipeline(
    call_tool,
    delete_item_if_exists,
    notebook_fixture_path,
    workspace_name,
):
    pipeline_name = unique_name("e2e_pipeline_notebook")
    notebook_name = unique_name("e2e_notebook")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline notebook activity",
        )
        assert create_result["status"] == "success"

        notebook_content = json.loads(notebook_fixture_path.read_text())
        import_result = await call_tool(
            "create_notebook",
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            notebook_content=notebook_content,
        )
        assert import_result["status"] == "success"

        add_result = await call_tool(
            "add_notebook_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            notebook_name=notebook_name,
        )
        assert add_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")
        await delete_item_if_exists(notebook_name, "Notebook")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_add_dataflow_activity_to_pipeline(
    call_tool,
    delete_item_if_exists,
    dataflow_name,
    workspace_name,
):
    pipeline_name = unique_name("e2e_pipeline_dataflow")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline dataflow activity",
        )
        assert create_result["status"] == "success"

        add_result = await call_tool(
            "add_dataflow_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            dataflow_name=dataflow_name,
        )
        assert add_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_add_activity_dependency_tool(
    call_tool,
    delete_item_if_exists,
    workspace_name,
):
    pipeline_name = unique_name("e2e_pipeline_dependency")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline dependency",
        )
        assert create_result["status"] == "success"

        activity_a = {
            "name": "WaitA",
            "type": "Wait",
            "dependsOn": [],
            "typeProperties": {"waitTimeInSeconds": 1},
        }
        activity_b = {
            "name": "WaitB",
            "type": "Wait",
            "dependsOn": [],
            "typeProperties": {"waitTimeInSeconds": 1},
        }

        add_a = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity_a,
        )
        assert add_a["status"] == "success"

        add_b = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity_b,
        )
        assert add_b["status"] == "success"

        dependency_result = await call_tool(
            "add_activity_dependency",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_name="WaitB",
            depends_on=["WaitA"],
        )
        assert dependency_result["status"] == "success"
        assert dependency_result.get("added_count", 0) >= 1

        definition = await call_tool(
            "get_pipeline_definition",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
        )
        assert definition["status"] == "success"
        activities = (
            definition.get("pipeline_content_json", {})
            .get("properties", {})
            .get("activities", [])
        )
        target = next(
            (activity for activity in activities if activity.get("name") == "WaitB"),
            None,
        )
        assert target is not None
        depends_on = target.get("dependsOn", [])
        assert any(dep.get("activity") == "WaitA" for dep in depends_on)
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_remove_dependencies_and_delete_activity(
    call_tool,
    delete_item_if_exists,
    workspace_name,
):
    pipeline_name = unique_name("e2e_pipeline_delete_dep")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline delete activity",
        )
        assert create_result["status"] == "success"

        activity_a = {
            "name": "WaitA",
            "type": "Wait",
            "dependsOn": [],
            "typeProperties": {"waitTimeInSeconds": 1},
        }
        activity_b = {
            "name": "WaitB",
            "type": "Wait",
            "dependsOn": [{"activity": "WaitA", "dependencyConditions": ["Succeeded"]}],
            "typeProperties": {"waitTimeInSeconds": 1},
        }
        activity_c = {
            "name": "WaitC",
            "type": "Wait",
            "dependsOn": [{"activity": "WaitA", "dependencyConditions": ["Succeeded"]}],
            "typeProperties": {"waitTimeInSeconds": 1},
        }

        add_a = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity_a,
        )
        assert add_a["status"] == "success"

        add_b = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity_b,
        )
        assert add_b["status"] == "success"

        add_c = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity_c,
        )
        assert add_c["status"] == "success"

        delete_blocked = await call_tool(
            "delete_activity_from_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_name="WaitA",
        )
        assert delete_blocked["status"] == "error"

        remove_result = await call_tool(
            "remove_activity_dependency",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_name="WaitA",
        )
        assert remove_result["status"] == "success"
        assert remove_result["removed_count"] == 2

        delete_result = await call_tool(
            "delete_activity_from_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_name="WaitA",
        )
        assert delete_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_remove_dependency_from_specific_activity(
    call_tool,
    delete_item_if_exists,
    workspace_name,
):
    pipeline_name = unique_name("e2e_pipeline_delete_dep_specific")
    try:
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline delete activity (specific)",
        )
        assert create_result["status"] == "success"

        activity_a = {
            "name": "WaitA",
            "type": "Wait",
            "dependsOn": [],
            "typeProperties": {"waitTimeInSeconds": 1},
        }
        activity_b = {
            "name": "WaitB",
            "type": "Wait",
            "dependsOn": [{"activity": "WaitA", "dependencyConditions": ["Succeeded"]}],
            "typeProperties": {"waitTimeInSeconds": 1},
        }

        add_a = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity_a,
        )
        assert add_a["status"] == "success"

        add_b = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity_b,
        )
        assert add_b["status"] == "success"

        remove_result = await call_tool(
            "remove_activity_dependency",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_name="WaitA",
            from_activity_name="WaitB",
        )
        assert remove_result["status"] == "success"
        assert remove_result["removed_count"] == 1

        delete_result = await call_tool(
            "delete_activity_from_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_name="WaitA",
        )
        assert delete_result["status"] == "success"
    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")


@pytest.mark.integration
@pytest.mark.asyncio
async def test_get_pipeline_activity_runs(
    call_tool,
    delete_item_if_exists,
    workspace_name,
):
    """Test get_pipeline_activity_runs after running a pipeline."""
    pipeline_name = unique_name("e2e_pipeline_activity_runs")
    try:
        # Create pipeline with a simple Wait activity
        create_result = await call_tool(
            "create_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            description="Integration test pipeline for activity runs",
        )
        assert create_result["status"] == "success"

        activity = {
            "name": "WaitActivity",
            "type": "Wait",
            "dependsOn": [],
            "typeProperties": {"waitTimeInSeconds": 1},
        }
        add_result = await call_tool(
            "add_activity_to_pipeline",
            workspace_name=workspace_name,
            pipeline_name=pipeline_name,
            activity_json=activity,
        )
        assert add_result["status"] == "success"

        # Run the pipeline
        run_result = await call_tool(
            "run_on_demand_job",
            workspace_name=workspace_name,
            item_name=pipeline_name,
            item_type="DataPipeline",
            job_type="Pipeline",
        )
        assert run_result["status"] == "success"
        job_instance_id = run_result["job_instance_id"]

        # Wait for pipeline to complete (poll status)
        max_wait = 120  # 2 minutes max
        poll_interval = 5
        elapsed = 0
        while elapsed < max_wait:
            status_result = await call_tool(
                "get_job_status",
                workspace_name=workspace_name,
                item_name=pipeline_name,
                item_type="DataPipeline",
                job_instance_id=job_instance_id,
            )
            assert status_result["status"] == "success"
            job = status_result.get("job", {})
            if job.get("is_terminal"):
                break
            await asyncio.sleep(poll_interval)
            elapsed += poll_interval

        # Wait a bit more for activity runs to be available
        await asyncio.sleep(5)

        # Now get activity runs
        activity_runs_result = await call_tool(
            "get_pipeline_activity_runs",
            workspace_name=workspace_name,
            job_instance_id=job_instance_id,
        )
        assert activity_runs_result["status"] == "success"

        # Activity runs may not be immediately available; check gracefully
        activity_count = activity_runs_result["activity_count"]
        if activity_count == 0:
            # Retry once after additional delay
            await asyncio.sleep(10)
            activity_runs_result = await call_tool(
                "get_pipeline_activity_runs",
                workspace_name=workspace_name,
                job_instance_id=job_instance_id,
            )
            activity_count = activity_runs_result["activity_count"]

        assert (
            activity_count >= 1
        ), f"Expected at least 1 activity run, got {activity_count}"
        # pipeline_name may be None if not in the activity runs response
        # This is acceptable as long as we get activities back

        # Check that WaitActivity is in the results
        activities = activity_runs_result.get("activities", [])
        activity_names = [a["activity_name"] for a in activities]
        assert "WaitActivity" in activity_names

        # Verify activity details
        wait_activity = next(
            a for a in activities if a["activity_name"] == "WaitActivity"
        )
        assert wait_activity["activity_type"] == "Wait"
        assert wait_activity["status"] in [
            "Succeeded",
            "Failed",
            "InProgress",
            "Cancelled",
        ]
        assert (
            wait_activity["duration_ms"] is not None
            or wait_activity["status"] == "InProgress"
        )

    finally:
        await delete_item_if_exists(pipeline_name, "DataPipeline")



# ABOUTME: SQL MCP tools for Microsoft Fabric.
# ABOUTME: Provides tools for SQL endpoint discovery and query execution.
"""SQL warehouse MCP tools.

This module provides MCP tools for executing SQL queries and DML statements
against Microsoft Fabric SQL warehouses.
"""

from typing import TYPE_CHECKING
import logging

if TYPE_CHECKING:
    from fastmcp import FastMCP

from ..services import FabricSQLService
from .base import handle_tool_errors, log_tool_invocation

logger = logging.getLogger(__name__)


def register_sql_tools(mcp: "FastMCP", sql_service: FabricSQLService):
    """Register SQL warehouse MCP tools.
    
    This function registers SQL-related tools:
    - get_sql_endpoint: Get SQL connection string for a Warehouse or Lakehouse (unified)
    - execute_sql_query: Execute SELECT/SHOW/DESCRIBE queries
    - execute_sql_statement: Execute DML statements (INSERT/UPDATE/DELETE)
    
    Args:
        mcp: FastMCP server instance to register tools on.
        sql_service: Initialized FabricSQLService instance.
        
    Example:
        ```python
        from ms_fabric_mcp_server import (
            FabricConfig, FabricClient,
            FabricWorkspaceService, FabricItemService, FabricSQLService
        )
        from ms_fabric_mcp_server.tools import register_sql_tools
        
        config = FabricConfig.from_environment()
        client = FabricClient(config)
        workspace_service = FabricWorkspaceService(client)
        item_service = FabricItemService(client)
        sql_service = FabricSQLService(client, workspace_service, item_service)
        
        register_sql_tools(mcp, sql_service)
        ```
    """
    
    @mcp.tool(title="Get SQL Endpoint")
    @handle_tool_errors
    def get_sql_endpoint(
        workspace_name: str,
        item_name: str,
        item_type: str = "Warehouse"
    ) -> dict:
        """Get the SQL endpoint/connection string for a Fabric Warehouse or Lakehouse.
        
        This unified tool supports both Warehouse and Lakehouse SQL endpoints, enabling
        direct T-SQL queries against either item type using the same connection pattern.
        
        Parameters:
            workspace_name: The display name of the workspace containing the item.
            item_name: The name of the Warehouse or Lakehouse to get the SQL endpoint for.
            item_type: Type of the item: "Warehouse" (default) or "Lakehouse".
                      Both types support T-SQL queries via ODBC connections.
            
        Returns:
            Dictionary with status, workspace_name, item_name, item_type, item_id,
            and connection_string (SQL endpoint URL).
            
        Example:
            ```python
            # Get Warehouse SQL endpoint
            result = get_sql_endpoint(
                workspace_name="My Workspace",
                item_name="Sales_Warehouse",
                item_type="Warehouse"
            )
            
            # Get Lakehouse SQL endpoint
            result = get_sql_endpoint(
                workspace_name="My Workspace",
                item_name="Data_Lakehouse",
                item_type="Lakehouse"
            )
            
            if result["status"] == "success":
                endpoint = result["connection_string"]
                # Use endpoint with execute_sql_query or external tools
            ```
        """
        log_tool_invocation("get_sql_endpoint",
                          workspace_name=workspace_name, item_name=item_name, item_type=item_type)
        logger.info(f"Getting SQL endpoint for {item_type} '{item_name}' in workspace '{workspace_name}'")
        
        connection_string = sql_service.get_sql_endpoint(workspace_name, item_name, item_type)
        
        result = {
            "status": "success",
            "workspace_name": workspace_name,
            "item_name": item_name,
            "item_type": item_type,
            "connection_string": connection_string,
            "message": f"Successfully retrieved connection string for {item_type} '{item_name}'"
        }
        
        logger.info(f"Successfully retrieved connection string for {item_type} '{item_name}'")
        return result

    @mcp.tool(title="Execute SQL Query")
    @handle_tool_errors
    def execute_sql_query(
        sql_endpoint: str,
        query: str,
        database: str = "Metadata"
    ) -> dict:
        """Execute a SQL query against a Fabric SQL Warehouse.
        
        Executes a SELECT, SHOW, or DESCRIBE query and returns the results as structured data.
        The query results include column names, data rows, and row count.
        
        **Note**: This tool is for queries that return data (SELECT/SHOW/DESCRIBE).
        For INSERT/UPDATE/DELETE statements, use execute_sql_statement instead.
        
        Parameters:
            sql_endpoint: The SQL endpoint URL (e.g., "abc-123.datawarehouse.fabric.microsoft.com").
                         Use get_sql_endpoint to retrieve this.
            query: SQL query to execute (SELECT, SHOW, DESCRIBE).
            database: Database/Warehouse name to connect to (default: "Metadata").
            
        Returns:
            Dictionary with status, message, data (rows), columns, row_count, and endpoint.
            
        Example:
            ```python
            # Get the endpoint first
            endpoint_result = get_sql_endpoint("My Workspace", "Sales_Warehouse", "Warehouse")
            endpoint = endpoint_result["connection_string"]
            
            # Execute a query
            result = execute_sql_query(
                sql_endpoint=endpoint,
                query="SELECT TOP 10 * FROM sales_data",
                database="sales_db"
            )
            
            if result["status"] == "success":
                for row in result["data"]:
                    print(row)
            ```
        """
        log_tool_invocation("execute_sql_query",
                          sql_endpoint=sql_endpoint, query=query[:100], database=database)
        logger.info(f"Executing SQL query on {sql_endpoint} (database: {database}): {query[:100]}...")
        
        result = sql_service.execute_sql_query(
            sql_endpoint=sql_endpoint,
            query=query,
            database=database
        )
        
        response = {
            "status": result.status,
            "message": result.message,
            "data": result.data,
            "columns": result.columns,
            "row_count": result.row_count,
            "endpoint": sql_endpoint
        }
        
        if result.status == "success":
            logger.info(f"Query executed successfully: {result.row_count} rows returned")
        else:
            logger.error(f"Query execution failed: {result.message}")
            
        return response

    @mcp.tool(title="Execute SQL DML Statement")
    @handle_tool_errors
    def execute_sql_statement(
        sql_endpoint: str,
        statement: str,
        database: str = "Metadata",
        allow_ddl: bool = False,
    ) -> dict:
        """Execute a DML SQL statement (INSERT, UPDATE, DELETE, MERGE) against a Fabric SQL Warehouse.
        
        Executes a data modification statement and returns the number of affected rows.
        
        **Note**: This tool is for DML statements that modify data (INSERT/UPDATE/DELETE/MERGE).
        For queries that return data (SELECT/SHOW/DESCRIBE), use execute_sql_query instead.
        Use allow_ddl=True to permit DDL statements (CREATE/ALTER/DROP/TRUNCATE). DDL is
        supported only on Warehouse SQL endpoints; Lakehouse SQL endpoints are read-only.
        
        Parameters:
            sql_endpoint: The SQL endpoint URL (e.g., "abc-123.datawarehouse.fabric.microsoft.com").
                         Use get_sql_endpoint to retrieve this.
            statement: DML SQL statement to execute (INSERT, UPDATE, DELETE, MERGE).
            database: Database/Warehouse name to connect to (default: "Metadata").
            allow_ddl: If True, allow DDL statements (CREATE/ALTER/DROP/TRUNCATE).
            
        Returns:
            Dictionary with status, message, affected_rows, and endpoint.
            
        Example:
            ```python
            # Get the endpoint first
            endpoint_result = get_sql_endpoint("My Workspace", "Sales_Warehouse", "Warehouse")
            endpoint = endpoint_result["connection_string"]
            
            # Execute an INSERT statement
            result = execute_sql_statement(
                sql_endpoint=endpoint,
                statement="INSERT INTO sales_data (id, amount) VALUES (1, 100.00)",
                database="sales_db"
            )
            
            if result["status"] == "success":
                print(f"Inserted {result['affected_rows']} row(s)")
            ```
        """
        log_tool_invocation("execute_sql_statement",
                          sql_endpoint=sql_endpoint, statement=statement[:100], database=database,
                          allow_ddl=allow_ddl)
        logger.info(f"Executing DML statement on {sql_endpoint} (database: {database}): {statement[:100]}...")
        
        result = sql_service.execute_sql_statement(
            sql_endpoint=sql_endpoint,
            statement=statement,
            database=database,
            allow_ddl=allow_ddl,
        )
        
        response = {
            "status": result["status"],
            "message": result["message"],
            "affected_rows": result["affected_rows"],
            "endpoint": sql_endpoint
        }
        
        if result["status"] == "success":
            logger.info(f"DML statement executed successfully: {result['affected_rows']} rows affected")
        else:
            logger.error(f"DML statement execution failed: {result['message']}")
            
        return response
    
    logger.info("SQL tools registered successfully (3 tools)")



"""Unit tests for FabricNotebookService."""

import base64
import json
from unittest.mock import Mock, patch

import pytest

from ms_fabric_mcp_server.client.exceptions import (
    FabricAPIError,
    FabricError,
    FabricItemNotFoundError,
    FabricValidationError,
)
from ms_fabric_mcp_server.models.item import FabricItem
from ms_fabric_mcp_server.models.job import FabricJob
from ms_fabric_mcp_server.models.results import JobStatusResult
from tests.fixtures.mocks import MockResponseFactory


def _make_response(status_code=200, json_data=None, text="", headers=None):
    response = Mock()
    response.status_code = status_code
    response.ok = 200 <= status_code < 300
    response.json.return_value = json_data or {}
    response.text = text
    response.headers = headers or {}
    return response


def _encode_ipynb(payload):
    return base64.b64encode(json.dumps(payload).encode("utf-8")).decode("utf-8")


@pytest.fixture
def mock_workspace_service():
    service = Mock()
    service.resolve_workspace_id.return_value = "workspace-123"
    return service


@pytest.fixture
def mock_item_service():
    return Mock()


@pytest.fixture
def notebook_service(mock_fabric_client, mock_item_service, mock_workspace_service):
    from ms_fabric_mcp_server.services.notebook import FabricNotebookService

    return FabricNotebookService(
        mock_fabric_client,
        mock_item_service,
        mock_workspace_service,
    )


@pytest.mark.unit
class TestFabricNotebookService:
    """Test suite for FabricNotebookService."""

    def test_encode_notebook_content_success(self, notebook_service):
        """Encode a notebook dict to base64 JSON."""
        payload = {"cells": [], "metadata": {}, "nbformat": 4, "nbformat_minor": 5}
        encoded = notebook_service._encode_notebook_content(payload)

        decoded = base64.b64decode(encoded.encode("utf-8")).decode("utf-8")
        assert json.loads(decoded) == payload

    def test_encode_notebook_content_invalid(self, notebook_service):
        """Invalid notebook content raises validation error."""
        with pytest.raises(FabricValidationError):
            notebook_service._encode_notebook_content({})

    def test_create_notebook_definition_includes_description(self, notebook_service):
        """Notebook definition includes description when provided."""
        notebook_service._encode_notebook_content = Mock(return_value="encoded")

        definition = notebook_service._create_notebook_definition(
            notebook_name="Test Notebook",
            notebook_content={"cells": []},
            description="Test description",
        )

        assert definition["displayName"] == "Test Notebook"
        assert definition["type"] == "Notebook"
        assert definition["description"] == "Test description"
        part = definition["definition"]["parts"][0]
        assert part["path"] == "Test Notebook.ipynb"
        assert part["payload"] == "encoded"

    def test_create_notebook_definition_without_description(self, notebook_service):
        """Notebook definition omits description when not provided."""
        notebook_service._encode_notebook_content = Mock(return_value="encoded")

        definition = notebook_service._create_notebook_definition(
            notebook_name="Test Notebook",
            notebook_content={"cells": []},
        )

        assert "description" not in definition

    def test_create_notebook_definition_with_folder(self, notebook_service):
        """Notebook definition includes folderId when provided."""
        notebook_service._encode_notebook_content = Mock(return_value="encoded")

        definition = notebook_service._create_notebook_definition(
            notebook_name="Test Notebook",
            notebook_content={"cells": []},
            folder_id="folder-1",
        )

        assert definition["folderId"] == "folder-1"

    def test_create_notebook_success(self, notebook_service, mock_item_service, mock_workspace_service):
        """Create notebook returns success result."""
        notebook_service._create_notebook_definition = Mock(return_value={"definition": {}})
        mock_workspace_service.resolve_workspace_id.return_value = "ws-1"
        mock_item_service.resolve_folder_id_from_path.return_value = "folder-1"
        mock_item_service.create_item.return_value = FabricItem(
            id="nb-123",
            display_name="Test",
            type="Notebook",
            workspace_id="ws-1",
        )

        result = notebook_service.create_notebook(
            workspace_name="Workspace",
            notebook_name="Notebook",
            notebook_content={"cells": []},
            folder_path="Notebooks/Finance",
        )

        assert result.status == "success"
        assert result.notebook_id == "nb-123"
        notebook_service._create_notebook_definition.assert_called_once_with(
            "Notebook",
            {"cells": []},
            None,
            folder_id="folder-1",
        )
        mock_item_service.create_item.assert_called_once()

    @pytest.mark.parametrize(
        "exception",
        [
            FabricItemNotFoundError("Notebook", "Notebook", "Workspace"),
            FabricValidationError("field", "value", "bad"),
            FabricAPIError(400, "bad"),
        ],
    )
    def test_create_notebook_expected_errors(self, notebook_service, exception):
        """Create notebook maps known exceptions to error result."""
        notebook_service._create_notebook_definition = Mock(side_effect=exception)

        result = notebook_service.create_notebook(
            workspace_name="Workspace",
            notebook_name="Notebook",
            notebook_content={"cells": []},
        )

        assert result.status == "error"
        assert result.message

    def test_create_notebook_invalid_name(self, notebook_service):
        """Notebook names with separators raise validation error."""
        with pytest.raises(FabricValidationError):
            notebook_service.create_notebook(
                workspace_name="Workspace",
                notebook_name="Bad/Name",
                notebook_content={"cells": []},
            )

    def test_create_notebook_unexpected_error(self, notebook_service):
        """Unexpected exceptions return error with prefix."""
        notebook_service._create_notebook_definition = Mock(side_effect=RuntimeError("boom"))

        result = notebook_service.create_notebook(
            workspace_name="Workspace",
            notebook_name="Notebook",
            notebook_content={"cells": []},
        )

        assert result.status == "error"
        assert "Unexpected error" in result.message

    def test_get_notebook_definition_success(self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client):
        """Get notebook content decodes ipynb payload."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        notebook_content = {"cells": [{"cell_type": "code", "source": ["print(1)"]}]}
        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook.ipynb",
                        "payload": _encode_ipynb(notebook_content),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }

        response = _make_response(200, definition_response)
        mock_fabric_client.make_api_request.return_value = response

        result = notebook_service.get_notebook_definition("Workspace", "Notebook")

        assert result == notebook_content
        mock_fabric_client.make_api_request.assert_called_once()

    def test_get_notebook_definition_lro_success(self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client):
        """Handle 202 LRO with success status."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        notebook_content = {"cells": [{"cell_type": "markdown", "source": ["hi"]}]}
        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook.ipynb",
                        "payload": _encode_ipynb(notebook_content),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }

        initial = _make_response(202, headers={"Location": "https://poll", "Retry-After": "0"})
        poll = _make_response(200, {"status": "Succeeded"})
        result_resp = _make_response(200, definition_response)

        mock_fabric_client.make_api_request.side_effect = [initial, poll, result_resp]

        with patch("time.sleep", return_value=None):
            result = notebook_service.get_notebook_definition("Workspace", "Notebook")

        assert result == notebook_content

    def test_get_notebook_definition_lro_failed(self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client):
        """LRO failed status raises FabricError."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        initial = _make_response(202, headers={"Location": "https://poll", "Retry-After": "0"})
        poll = _make_response(200, {"status": "Failed", "error": {"message": "boom"}})
        mock_fabric_client.make_api_request.side_effect = [initial, poll]

        with patch("time.sleep", return_value=None):
            with pytest.raises(FabricError):
                notebook_service.get_notebook_definition("Workspace", "Notebook")

    def test_get_notebook_definition_lro_retry_after_non_integer(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """LRO handles non-integer Retry-After headers."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        notebook_content = {"cells": [{"cell_type": "markdown", "source": ["hi"]}]}
        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook.ipynb",
                        "payload": _encode_ipynb(notebook_content),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }

        initial = _make_response(
            202,
            headers={"Location": "https://poll", "Retry-After": "Wed, 21 Oct 2015 07:28:00 GMT"},
        )
        poll = _make_response(200, {"status": "Succeeded"})
        result_resp = _make_response(200, definition_response)
        mock_fabric_client.make_api_request.side_effect = [initial, poll, result_resp]

        with patch("time.sleep", return_value=None) as sleep_mock:
            result = notebook_service.get_notebook_definition("Workspace", "Notebook")

        assert result == notebook_content
        sleep_mock.assert_called_once_with(5)

    def test_get_notebook_definition_lro_in_progress(self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client):
        """LRO continues polling on non-terminal status."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        notebook_content = {"cells": [{"cell_type": "markdown", "source": ["hi"]}]}
        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook.ipynb",
                        "payload": _encode_ipynb(notebook_content),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }

        initial = _make_response(202, headers={"Location": "https://poll", "Retry-After": "0"})
        poll_running = _make_response(200, {"status": "Running"}, headers={"Retry-After": "0"})
        poll_accepted = _make_response(202, headers={"Retry-After": "0"})
        poll_succeeded = _make_response(200, {"status": "Succeeded"})
        result_resp = _make_response(200, definition_response)

        mock_fabric_client.make_api_request.side_effect = [
            initial,
            poll_running,
            poll_accepted,
            poll_succeeded,
            result_resp,
        ]

        with patch("time.sleep", return_value=None):
            result = notebook_service.get_notebook_definition("Workspace", "Notebook")

        assert result == notebook_content

    def test_get_notebook_definition_lro_timeout(self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client):
        """LRO timeout raises FabricError after max retries."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        initial = _make_response(202, headers={"Location": "https://poll", "Retry-After": "0"})
        poll = _make_response(202, headers={"Retry-After": "0"})
        mock_fabric_client.make_api_request.side_effect = [initial] + [poll] * 30

        with patch("time.sleep", return_value=None):
            with pytest.raises(FabricError):
                notebook_service.get_notebook_definition("Workspace", "Notebook")

    def test_get_notebook_definition_lro_missing_location(self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client):
        """LRO without Location header raises FabricError."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        initial = _make_response(202)
        mock_fabric_client.make_api_request.return_value = initial

        with pytest.raises(FabricError):
            notebook_service.get_notebook_definition("Workspace", "Notebook")

    def test_get_notebook_definition_returns_raw_definition(self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client):
        """Return raw definition when no ipynb part exists."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        mock_item_service.get_item_by_name.return_value = FabricItem(
            id="nb-123",
            display_name="Notebook",
            type="Notebook",
            workspace_id="ws-123",
        )

        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook-content.py",
                        "payload": base64.b64encode(b"print(1)").decode("utf-8"),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }
        response = _make_response(200, definition_response)
        mock_fabric_client.make_api_request.return_value = response

        result = notebook_service.get_notebook_definition("Workspace", "Notebook")

        assert result == definition_response

    def test_list_notebooks(self, notebook_service, mock_item_service, mock_workspace_service):
        """List notebooks delegates to item service."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        items = [
            FabricItem(id="nb-1", display_name="A", type="Notebook", workspace_id="ws-123"),
            FabricItem(id="nb-2", display_name="B", type="Notebook", workspace_id="ws-123"),
        ]
        mock_item_service.list_items.return_value = items

        result = notebook_service.list_notebooks("Workspace")

        assert result == items
        mock_item_service.list_items.assert_called_once_with("ws-123", "Notebook")

    def test_get_notebook_by_name(self, notebook_service, mock_item_service, mock_workspace_service):
        """Get notebook by name delegates to item service."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        item = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="ws-123")
        mock_item_service.get_item_by_name.return_value = item

        result = notebook_service.get_notebook_by_name("Workspace", "Notebook")

        assert result == item
        mock_item_service.get_item_by_name.assert_called_once_with("ws-123", "Notebook", "Notebook")

    def test_update_notebook_metadata(self, notebook_service, mock_item_service, mock_workspace_service):
        """Update notebook metadata delegates to item service."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="ws-123")
        updated = FabricItem(id="nb-1", display_name="Notebook v2", type="Notebook", workspace_id="ws-123")
        mock_item_service.get_item_by_name.return_value = notebook
        mock_item_service.update_item.return_value = updated

        result = notebook_service.update_notebook_metadata("Workspace", "Notebook", {"displayName": "Notebook v2"})

        assert result == updated
        mock_item_service.update_item.assert_called_once_with("ws-123", "nb-1", {"displayName": "Notebook v2"})

    def test_delete_notebook(self, notebook_service, mock_item_service, mock_workspace_service):
        """Delete notebook delegates to item service."""
        mock_workspace_service.resolve_workspace_id.return_value = "ws-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="ws-123")
        mock_item_service.get_item_by_name.return_value = notebook

        notebook_service.delete_notebook("Workspace", "Notebook")

        mock_item_service.delete_item.assert_called_once_with("ws-123", "nb-1")

    def test_execute_notebook_success(self, notebook_service, mock_workspace_service, mock_item_service):
        """Execute notebook returns mapped ExecuteNotebookResult."""
        job = FabricJob(
            job_instance_id="job-1",
            item_id="nb-1",
            job_type="RunNotebook",
            status="Completed",
            invoke_type="Manual",
            root_activity_id="root-1",
            start_time_utc="2025-01-01T00:00:00Z",
            end_time_utc="2025-01-01T00:05:00Z",
        )
        job_result = JobStatusResult(status="success", job=job, message="ok")

        with patch("ms_fabric_mcp_server.services.job.FabricJobService") as mock_job_service:
            mock_job_service.return_value.run_notebook_job.return_value = job_result
            result = notebook_service.execute_notebook("Workspace", "Notebook", parameters={"a": 1})

        assert result.status == "success"
        assert result.job_instance_id == "job-1"
        assert result.job_status == "Completed"

    def test_execute_notebook_error_result(self, notebook_service):
        """Execute notebook returns error when job result is error."""
        job_result = JobStatusResult(status="error", message="boom")

        with patch("ms_fabric_mcp_server.services.job.FabricJobService") as mock_job_service:
            mock_job_service.return_value.run_notebook_job.return_value = job_result
            result = notebook_service.execute_notebook("Workspace", "Notebook")

        assert result.status == "error"
        assert result.message == "boom"

    def test_execute_notebook_wait_false_success(self, notebook_service):
        """Execute notebook succeeds when wait=False returns job metadata."""
        job_result = JobStatusResult(status="success", job_instance_id="job-2", message="started")

        with patch("ms_fabric_mcp_server.services.job.FabricJobService") as mock_job_service:
            mock_job_service.return_value.run_notebook_job.return_value = job_result
            result = notebook_service.execute_notebook("Workspace", "Notebook", wait=False)

        assert result.status == "success"
        assert result.job_instance_id == "job-2"

    def test_execute_notebook_unexpected_error(self, notebook_service):
        """Execute notebook wraps unexpected exceptions."""
        with patch("ms_fabric_mcp_server.services.job.FabricJobService") as mock_job_service:
            mock_job_service.return_value.run_notebook_job.side_effect = RuntimeError("boom")
            result = notebook_service.execute_notebook("Workspace", "Notebook")

        assert result.status == "error"
        assert "Unexpected error" in result.message

    def test_update_notebook_definition_with_lakehouse(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """Update notebook definition applies lakehouse dependencies when provided."""
        mock_workspace_service.resolve_workspace_id.side_effect = ["workspace-123", "workspace-123"]

        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        lakehouse = FabricItem(id="lh-1", display_name="Lakehouse", type="Lakehouse", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.side_effect = [notebook, lakehouse]

        existing_payload = {"cells": [], "metadata": {}}
        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook.ipynb",
                        "payload": _encode_ipynb(existing_payload),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }

        get_def = _make_response(200, definition_response)
        update_def = MockResponseFactory.success({})
        mock_fabric_client.make_api_request.side_effect = [get_def, update_def]

        result = notebook_service.update_notebook_definition(
            workspace_name="Workspace",
            notebook_name="Notebook",
            notebook_content={"cells": []},
            default_lakehouse_name="Lakehouse",
        )

        assert result.status == "success"
        update_call = mock_fabric_client.make_api_request.call_args_list[1]
        payload = update_call.kwargs["payload"]
        encoded = payload["definition"]["parts"][0]["payload"]
        decoded = json.loads(base64.b64decode(encoded).decode("utf-8"))
        lakehouse_meta = decoded["metadata"]["dependencies"]["lakehouse"]
        assert lakehouse_meta["default_lakehouse"] == "lh-1"
        assert lakehouse_meta["default_lakehouse_name"] == "Lakehouse"
        assert lakehouse_meta["default_lakehouse_workspace_id"] == "workspace-123"
        assert lakehouse_meta["known_lakehouses"] == [{"id": "lh-1"}]

    def test_update_notebook_definition_preserves_dependencies(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """Update notebook definition preserves existing dependencies when no lakehouse specified."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"

        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.return_value = notebook

        existing_payload = {"cells": [], "metadata": {"dependencies": {"foo": "bar"}}}
        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook.ipynb",
                        "payload": _encode_ipynb(existing_payload),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }

        get_def = _make_response(200, definition_response)
        update_def = MockResponseFactory.success({})
        mock_fabric_client.make_api_request.side_effect = [get_def, update_def]

        result = notebook_service.update_notebook_definition(
            workspace_name="Workspace",
            notebook_name="Notebook",
            notebook_content={"cells": []},
        )

        assert result.status == "success"
        update_call = mock_fabric_client.make_api_request.call_args_list[1]
        payload = update_call.kwargs["payload"]
        encoded = payload["definition"]["parts"][0]["payload"]
        decoded = json.loads(base64.b64decode(encoded).decode("utf-8"))
        assert decoded["metadata"]["dependencies"] == {"foo": "bar"}

    def test_update_notebook_definition_uses_existing_definition(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """Update notebook definition can reuse existing definition when content omitted."""
        mock_workspace_service.resolve_workspace_id.side_effect = ["workspace-123", "workspace-123"]

        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        lakehouse = FabricItem(id="lh-1", display_name="Lakehouse", type="Lakehouse", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.side_effect = [notebook, lakehouse]

        existing_payload = {"cells": [], "metadata": {"dependencies": {"foo": "bar"}}}
        definition_response = {
            "definition": {
                "parts": [
                    {
                        "path": "notebook.ipynb",
                        "payload": _encode_ipynb(existing_payload),
                        "payloadType": "InlineBase64",
                    }
                ]
            }
        }

        get_def = _make_response(200, definition_response)
        update_def = MockResponseFactory.success({})
        mock_fabric_client.make_api_request.side_effect = [get_def, update_def]

        result = notebook_service.update_notebook_definition(
            workspace_name="Workspace",
            notebook_name="Notebook",
            notebook_content=None,
            default_lakehouse_name="Lakehouse",
        )

        assert result.status == "success"
        update_call = mock_fabric_client.make_api_request.call_args_list[1]
        payload = update_call.kwargs["payload"]
        encoded = payload["definition"]["parts"][0]["payload"]
        decoded = json.loads(base64.b64decode(encoded).decode("utf-8"))
        assert decoded["metadata"]["dependencies"]["lakehouse"]["default_lakehouse"] == "lh-1"

    def test_update_notebook_definition_missing_lakehouse(
        self, notebook_service, mock_item_service, mock_workspace_service
    ):
        """Missing lakehouse returns error result."""
        mock_workspace_service.resolve_workspace_id.side_effect = ["workspace-123", "workspace-123"]

        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.side_effect = [
            notebook,
            FabricItemNotFoundError("Lakehouse", "Lakehouse", "workspace-123"),
        ]

        result = notebook_service.update_notebook_definition(
            workspace_name="Workspace",
            notebook_name="Notebook",
            notebook_content={"cells": []},
            default_lakehouse_name="Lakehouse",
        )

        assert result.status == "error"

    def test_get_notebook_run_details_success(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """Return execution details and summary."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.return_value = notebook

        list_response = _make_response(
            200,
            {
                "value": [
                    {
                        "jobInstanceId": "job-1",
                        "livyId": "livy-1",
                        "state": "Success",
                        "sparkApplicationId": "app-1",
                        "operationName": "RunNotebook",
                        "submittedDateTime": "2025-01-01T00:00:00Z",
                        "startDateTime": "2025-01-01T00:01:00Z",
                        "endDateTime": "2025-01-01T00:05:00Z",
                        "totalDuration": {"value": 240},
                    }
                ]
            },
        )
        detail_response = _make_response(
            200,
            {
                "state": "Success",
                "sparkApplicationId": "app-1",
                "livyId": "livy-1",
                "jobInstanceId": "job-1",
                "operationName": "RunNotebook",
                "submittedDateTime": "2025-01-01T00:00:00Z",
                "startDateTime": "2025-01-01T00:01:00Z",
                "endDateTime": "2025-01-01T00:05:00Z",
                "queuedDuration": {"value": 5},
                "runningDuration": {"value": 235},
                "totalDuration": {"value": 240},
            },
        )
        job_response = _make_response(200, {"failureReason": {"message": "none"}})

        mock_fabric_client.make_api_request.side_effect = [
            list_response,
            detail_response,
            job_response,
        ]

        result = notebook_service.get_notebook_run_details("Workspace", "Notebook", "job-1")

        assert result["status"] == "success"
        assert result["execution_summary"]["failure_reason"] == {"message": "none"}
        assert result["notebook_id"] == "nb-1"

    def test_get_notebook_run_details_no_session(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """Return error when no matching session found."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.return_value = notebook

        list_response = _make_response(200, {"value": []})
        mock_fabric_client.make_api_request.return_value = list_response

        result = notebook_service.get_notebook_run_details("Workspace", "Notebook", "job-1")

        assert result["status"] == "error"
        assert result["available_sessions"] == 0

    def test_get_notebook_run_details_api_error(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """API errors return error status."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.return_value = notebook

        mock_fabric_client.make_api_request.side_effect = FabricAPIError(500, "boom")

        result = notebook_service.get_notebook_run_details("Workspace", "Notebook", "job-1")

        assert result["status"] == "error"
        assert "boom" in result["message"]

    def test_list_notebook_runs_success_with_limit(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """List executions applies limit to session summaries."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.return_value = notebook

        list_response = _make_response(
            200,
            {
                "value": [
                    {"jobInstanceId": "job-1", "livyId": "1", "state": "Success"},
                    {"jobInstanceId": "job-2", "livyId": "2", "state": "Failed"},
                ]
            },
        )
        mock_fabric_client.make_api_request.return_value = list_response

        result = notebook_service.list_notebook_runs("Workspace", "Notebook", limit=1)

        assert result["status"] == "success"
        assert len(result["sessions"]) == 1
        assert result["total_count"] == 2

    def test_list_notebook_runs_item_not_found(
        self, notebook_service, mock_item_service, mock_workspace_service
    ):
        """Item not found maps to error response."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"
        mock_item_service.get_item_by_name.side_effect = FabricItemNotFoundError("Notebook", "Notebook", "workspace-123")

        result = notebook_service.list_notebook_runs("Workspace", "Notebook")

        assert result["status"] == "error"

    def test_get_notebook_driver_logs_success_truncated(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """Driver log retrieval truncates to max lines."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.return_value = notebook

        notebook_service.get_notebook_run_details = Mock(
            return_value={
                "status": "success",
                "execution_summary": {
                    "livy_id": "livy-1",
                    "spark_application_id": "app-1",
                },
            }
        )

        meta_response = _make_response(200, {"sizeInBytes": 123})
        log_response = _make_response(200, text="line1\nline2\nline3")
        mock_fabric_client.make_api_request.side_effect = [meta_response, log_response]

        result = notebook_service.get_notebook_driver_logs(
            "Workspace",
            "Notebook",
            "job-1",
            log_type="stdout",
            max_lines=2,
        )

        assert result["status"] == "success"
        assert result["truncated"] is True
        assert result["log_content"] == "line2\nline3"

    def test_get_notebook_driver_logs_invalid_log_type(self, notebook_service):
        """Invalid log_type returns error result."""
        result = notebook_service.get_notebook_driver_logs(
            "Workspace", "Notebook", "job-1", log_type="bad"
        )

        assert result["status"] == "error"
        assert "Invalid log_type" in result["message"]

    def test_get_notebook_driver_logs_exec_details_error(self, notebook_service):
        """Propagates execution detail error."""
        notebook_service.get_notebook_run_details = Mock(
            return_value={"status": "error", "message": "boom"}
        )

        result = notebook_service.get_notebook_driver_logs(
            "Workspace", "Notebook", "job-1"
        )

        assert result["status"] == "error"
        assert result["message"] == "boom"

    def test_get_notebook_driver_logs_missing_summary_fields(self, notebook_service):
        """Missing livy or spark app IDs returns error."""
        notebook_service.get_notebook_run_details = Mock(
            return_value={"status": "success", "execution_summary": {}}
        )

        result = notebook_service.get_notebook_driver_logs(
            "Workspace", "Notebook", "job-1"
        )

        assert result["status"] == "error"

    def test_get_notebook_driver_logs_api_error(
        self, notebook_service, mock_item_service, mock_workspace_service, mock_fabric_client
    ):
        """API errors are mapped to error result."""
        mock_workspace_service.resolve_workspace_id.return_value = "workspace-123"
        notebook = FabricItem(id="nb-1", display_name="Notebook", type="Notebook", workspace_id="workspace-123")
        mock_item_service.get_item_by_name.return_value = notebook
        notebook_service.get_notebook_run_details = Mock(
            return_value={
                "status": "success",
                "execution_summary": {"livy_id": "livy-1", "spark_application_id": "app-1"},
            }
        )
        mock_fabric_client.make_api_request.side_effect = FabricAPIError(500, "boom")

        result = notebook_service.get_notebook_driver_logs(
            "Workspace", "Notebook", "job-1"
        )

        assert result["status"] == "error"



"""Integration tests for lakehouse file tools."""

import pytest

from tests.conftest import unique_name


@pytest.mark.integration
@pytest.mark.asyncio
async def test_upload_list_delete_lakehouse_file(
    call_tool,
    workspace_name,
    delete_item_if_exists,
    poll_until,
    tmp_path,
):
    lakehouse_name = unique_name("e2e_lakehouse_files")
    file_name = unique_name("sample_file")
    destination_path = f"raw/{file_name}.csv"

    lakehouse_id = None
    try:
        create_result = await call_tool(
            "create_lakehouse",
            workspace_name=workspace_name,
            lakehouse_name=lakehouse_name,
        )
        assert create_result["status"] == "success"
        lakehouse_id = create_result.get("lakehouse_id")
        assert lakehouse_id

        local_file = tmp_path / f"{file_name}.csv"
        local_file.write_text("col1,col2\n1,2\n")

        upload_result = await call_tool(
            "upload_lakehouse_file",
            workspace_name=workspace_name,
            lakehouse_id=lakehouse_id,
            local_file_path=str(local_file),
            destination_path=destination_path,
            create_missing_directories=True,
        )
        assert upload_result["status"] == "success"

        async def _file_visible():
            list_result = await call_tool(
                "list_lakehouse_files",
                workspace_name=workspace_name,
                lakehouse_id=lakehouse_id,
                path="raw",
                recursive=True,
            )
            if list_result.get("status") != "success":
                return None
            for entry in list_result.get("files", []):
                name = entry.get("name") or ""
                if destination_path in name:
                    return list_result
            return None

        listed = await poll_until(
            _file_visible, timeout_seconds=180, interval_seconds=10
        )
        assert listed is not None

        delete_result = await call_tool(
            "delete_lakehouse_file",
            workspace_name=workspace_name,
            lakehouse_id=lakehouse_id,
            path=destination_path,
        )
        assert delete_result["status"] == "success"
    finally:
        await delete_item_if_exists(lakehouse_name, "Lakehouse")



"""Unit tests for FabricJobService."""

from unittest.mock import Mock, patch

import pytest

from ms_fabric_mcp_server.client.exceptions import FabricAPIError, FabricItemNotFoundError
from ms_fabric_mcp_server.models.item import FabricItem
from ms_fabric_mcp_server.models.job import FabricJob
from ms_fabric_mcp_server.models.results import JobStatusResult, RunJobResult


def _make_response(status_code=200, json_data=None, headers=None):
    response = Mock()
    response.status_code = status_code
    response.ok = 200 <= status_code < 300
    response.json.return_value = json_data or {}
    response.headers = headers or {}
    return response


@pytest.fixture
def mock_workspace_service():
    service = Mock()
    service.resolve_workspace_id.return_value = "ws-123"
    return service


@pytest.fixture
def mock_item_service():
    service = Mock()
    service.get_item_by_name.return_value = FabricItem(
        id="item-123",
        display_name="Notebook",
        type="Notebook",
        workspace_id="ws-123",
    )
    return service


@pytest.fixture
def job_service(mock_fabric_client, mock_workspace_service, mock_item_service):
    from ms_fabric_mcp_server.services.job import FabricJobService

    return FabricJobService(mock_fabric_client, mock_workspace_service, mock_item_service)


@pytest.mark.unit
class TestFabricJobService:
    """Test suite for FabricJobService."""

    def test_run_on_demand_job_success(self, job_service, mock_fabric_client):
        """Run on-demand job parses headers and job ID."""
        headers = {
            "Location": (
                "https://api.fabric.microsoft.com/v1/workspaces/ws-123/items/item-123/"
                "jobs/instances/job-999?api-version=2023-11-01"
            ),
            "Retry-After": "15",
        }
        response = _make_response(202, json_data={}, headers=headers)
        mock_fabric_client.make_api_request.return_value = response

        result = job_service.run_on_demand_job(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_type="RunNotebook",
        )

        assert result.status == "success"
        assert result.job_instance_id == "job-999"
        assert result.retry_after == 15

    @pytest.mark.parametrize(
        "exception",
        [
            FabricItemNotFoundError("Notebook", "Notebook", "Workspace"),
            FabricAPIError(500, "boom"),
        ],
    )
    def test_run_on_demand_job_expected_errors(self, job_service, mock_workspace_service, mock_item_service, exception):
        """Known errors return error result."""
        if isinstance(exception, FabricItemNotFoundError):
            mock_item_service.get_item_by_name.side_effect = exception
        else:
            job_service.client.make_api_request.side_effect = exception

        result = job_service.run_on_demand_job(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_type="RunNotebook",
        )

        assert result.status == "error"

    def test_run_on_demand_job_unexpected_error(self, job_service, mock_fabric_client):
        """Unexpected errors return error result with prefix."""
        mock_fabric_client.make_api_request.side_effect = RuntimeError("boom")

        result = job_service.run_on_demand_job(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_type="RunNotebook",
        )

        assert result.status == "error"
        assert "Unexpected error" in result.message

    def test_get_job_status_success(self, job_service, mock_fabric_client):
        """get_job_status maps response into FabricJob."""
        response = _make_response(
            200,
            json_data={
                "id": "job-1",
                "itemId": "item-123",
                "jobType": "RunNotebook",
                "status": "Completed",
                "invokeType": "Manual",
                "rootActivityId": "root-1",
                "startTimeUtc": "2025-01-01T00:00:00Z",
                "endTimeUtc": "2025-01-01T00:05:00Z",
                "failureReason": {"message": "none"},
            },
        )
        mock_fabric_client.make_api_request.return_value = response

        result = job_service.get_job_status(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_instance_id="job-1",
        )

        assert result.status == "success"
        assert result.job is not None
        assert result.job.job_instance_id == "job-1"
        assert result.job.status == "Completed"
        assert result.job.failure_reason == {"message": "none"}

    def test_get_job_status_api_error(self, job_service, mock_fabric_client):
        """API errors return error result."""
        mock_fabric_client.make_api_request.side_effect = FabricAPIError(500, "boom")

        result = job_service.get_job_status(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_instance_id="job-1",
        )

        assert result.status == "error"

    def test_get_job_status_by_url_success(self, job_service, mock_fabric_client):
        """get_job_status_by_url strips version prefix."""
        response = _make_response(200, json_data={"id": "job-1", "status": "Completed"})
        mock_fabric_client.make_api_request.return_value = response

        result = job_service.get_job_status_by_url(
            "https://api.fabric.microsoft.com/v1/workspaces/ws/items/item/jobs/instances/job-1"
        )

        assert result.status == "success"
        mock_fabric_client.make_api_request.assert_called_once_with(
            "GET", "workspaces/ws/items/item/jobs/instances/job-1"
        )

    @pytest.mark.parametrize(
        "location_url",
        [
            "not-a-url",
            "https://api.fabric.microsoft.com/v1",
            "https://api.fabric.microsoft.com/",
            "https://api.fabric.microsoft.com/v1/workspaces/ws/items/item/jobs/instances",
            "https://api.fabric.microsoft.com/v1/workspaces/ws",
        ],
    )
    def test_get_job_status_by_url_invalid(self, job_service, mock_fabric_client, location_url):
        """Invalid URLs return error results without calling API."""
        result = job_service.get_job_status_by_url(location_url)

        assert result.status == "error"
        assert "Invalid job status URL" in result.message
        mock_fabric_client.make_api_request.assert_not_called()

    def test_run_on_demand_job_missing_location(self, job_service, mock_fabric_client):
        """Missing Location header returns error result."""
        response = _make_response(202, json_data={}, headers={})
        mock_fabric_client.make_api_request.return_value = response

        result = job_service.run_on_demand_job(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_type="RunNotebook",
        )

        assert result.status == "error"
        assert "Location" in result.message

    def test_run_notebook_job_no_wait_returns_job_id(self, job_service):
        """run_notebook_job(wait=False) surfaces job instance metadata."""
        run_result = RunJobResult(
            status="success",
            job_instance_id="job-123",
            location_url="https://api.fabric.microsoft.com/v1/workspaces/ws/items/item/jobs/instances/job-123",
            retry_after=10,
            message="ok",
        )

        with patch.object(job_service, "run_on_demand_job", return_value=run_result):
            result = job_service.run_notebook_job(
                workspace_name="Workspace",
                notebook_name="Notebook",
                wait=False,
            )

        assert result.status == "success"
        assert result.job_instance_id == "job-123"
        assert result.location_url == run_result.location_url
        assert result.retry_after == 10

    def test_wait_for_job_completion_success(self, job_service):
        """Wait for completion stops on terminal status."""
        in_progress = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="InProgress",
            ),
        )
        completed = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="Completed",
                end_time_utc="2025-01-01T00:05:00Z",
            ),
        )
        job_service.get_job_status = Mock(side_effect=[in_progress, completed])

        with patch("time.sleep", return_value=None):
            result = job_service.wait_for_job_completion(
                workspace_name="Workspace",
                item_name="Notebook",
                item_type="Notebook",
                job_instance_id="job-1",
                poll_interval=0,
                timeout_minutes=1,
            )

        assert result.job is not None
        assert result.job.status == "Completed"

    def test_wait_for_job_completion_by_url_failed(self, job_service):
        """Failed job by URL returns terminal status with failure reason."""
        failed = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="Failed",
                end_time_utc="2025-01-01T00:05:00Z",
                failure_reason={"message": "boom"},
            ),
        )
        job_service.get_job_status_by_url = Mock(return_value=failed)

        result = job_service.wait_for_job_completion_by_url(
            location_url="https://api.fabric.microsoft.com/v1/...",
            poll_interval=0,
            timeout_minutes=1,
        )

        assert result.job is not None
        assert result.job.is_failed()
        assert result.job.failure_reason == {"message": "boom"}

    def test_wait_for_job_completion_failed(self, job_service):
        """Failed job returns terminal status with failure reason."""
        failed = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="Failed",
                end_time_utc="2025-01-01T00:05:00Z",
                failure_reason={"message": "boom"},
            ),
        )
        job_service.get_job_status = Mock(return_value=failed)

        result = job_service.wait_for_job_completion(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_instance_id="job-1",
            poll_interval=0,
            timeout_minutes=1,
        )

        assert result.job is not None
        assert result.job.is_failed()
        assert result.job.failure_reason == {"message": "boom"}

    def test_wait_for_job_completion_returns_error_result(self, job_service):
        """Error result during polling returns immediately."""
        error_result = JobStatusResult(status="error", message="boom")
        job_service.get_job_status = Mock(return_value=error_result)

        result = job_service.wait_for_job_completion(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_instance_id="job-1",
            poll_interval=0,
            timeout_minutes=1,
        )

        assert result.status == "error"
        assert result.message == "boom"

    def test_wait_for_job_completion_timeout(self, job_service):
        """Timeout path returns final status with timeout message."""
        final = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="InProgress",
            ),
        )
        job_service.get_job_status = Mock(return_value=final)

        result = job_service.wait_for_job_completion(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_instance_id="job-1",
            poll_interval=0,
            timeout_minutes=0,
        )

        assert "Timed out" in result.message

    def test_wait_for_job_completion_by_url_success(self, job_service):
        """Wait for completion by URL stops on terminal status."""
        in_progress = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="InProgress",
            ),
        )
        completed = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="Completed",
                end_time_utc="2025-01-01T00:05:00Z",
            ),
        )
        job_service.get_job_status_by_url = Mock(side_effect=[in_progress, completed])

        with patch("time.sleep", return_value=None):
            result = job_service.wait_for_job_completion_by_url(
                location_url="https://api.fabric.microsoft.com/v1/...",
                poll_interval=0,
                timeout_minutes=1,
            )

        assert result.job is not None
        assert result.job.status == "Completed"

    def test_wait_for_job_completion_terminal_without_end_time(self, job_service):
        """Terminal status returns even without end_time_utc."""
        completed = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="Completed",
                end_time_utc=None,
            ),
        )
        job_service.get_job_status = Mock(return_value=completed)

        result = job_service.wait_for_job_completion(
            workspace_name="Workspace",
            item_name="Notebook",
            item_type="Notebook",
            job_instance_id="job-1",
            poll_interval=0,
            timeout_minutes=1,
        )

        assert result.job is not None
        assert result.job.status == "Completed"

    def test_wait_for_job_completion_by_url_terminal_without_end_time(self, job_service):
        """Terminal status by URL returns even without end_time_utc."""
        completed = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="Completed",
                end_time_utc=None,
            ),
        )
        job_service.get_job_status_by_url = Mock(return_value=completed)

        result = job_service.wait_for_job_completion_by_url(
            location_url="https://api.fabric.microsoft.com/v1/...",
            poll_interval=0,
            timeout_minutes=1,
        )

        assert result.job is not None
        assert result.job.status == "Completed"
    def test_wait_for_job_completion_by_url_timeout(self, job_service):
        """Timeout path returns final status with timeout message."""
        final = JobStatusResult(
            status="success",
            job=FabricJob(
                job_instance_id="job-1",
                item_id="item-123",
                job_type="RunNotebook",
                status="InProgress",
            ),
        )
        job_service.get_job_status_by_url = Mock(return_value=final)

        result = job_service.wait_for_job_completion_by_url(
            location_url="https://api.fabric.microsoft.com/v1/...",
            poll_interval=0,
            timeout_minutes=0,
        )

        assert "Timed out" in result.message



"""Integration tests for workspace tools."""

import pytest


@pytest.mark.integration
@pytest.mark.asyncio
async def test_list_workspaces_includes_configured(call_tool, workspace_name):
    result = await call_tool("list_workspaces")

    assert result["status"] == "success"
    workspace_names = [ws.get("display_name") for ws in result.get("workspaces", [])]
    assert workspace_name in workspace_names



# ABOUTME: Pipeline-related data models for Microsoft Fabric.
# ABOUTME: Provides PipelineActivityRun model for activity execution details.
"""Pipeline-related data models for Microsoft Fabric."""

from typing import Optional

from pydantic import BaseModel, ConfigDict, Field


class PipelineActivityRun(BaseModel):
    """Pipeline activity run details.

    Represents execution details for a single activity within a pipeline run,
    including timing, status, and output metrics.

    Attributes:
        activity_name: Name of the activity
        activity_type: Type of activity (Copy, Notebook, etc.)
        status: Execution status (Succeeded, Failed, InProgress, Cancelled)
        duration_ms: Duration in milliseconds
        start_time: Activity start time in UTC (ISO 8601)
        end_time: Activity end time in UTC (ISO 8601)
        rows_read: Number of rows read (Copy activities only)
        rows_written: Number of rows written (Copy activities only)
        error_message: Error message if activity failed

    Example:
        ```python
        activity = PipelineActivityRun(
            activity_name="CopyProducts",
            activity_type="Copy",
            status="Succeeded",
            duration_ms=5230,
            start_time="2026-02-03T13:00:00Z",
            end_time="2026-02-03T13:00:05Z",
            rows_read=1234,
            rows_written=1234
        )
        ```
    """

    activity_name: str = Field(description="Name of the activity")
    activity_type: str = Field(description="Type of activity (Copy, Notebook, etc.)")
    status: str = Field(description="Execution status")
    duration_ms: Optional[int] = Field(
        default=None, description="Duration in milliseconds"
    )
    start_time: Optional[str] = Field(
        default=None, description="Activity start time (UTC)"
    )
    end_time: Optional[str] = Field(default=None, description="Activity end time (UTC)")
    rows_read: Optional[int] = Field(default=None, description="Rows read (Copy only)")
    rows_written: Optional[int] = Field(
        default=None, description="Rows written (Copy only)"
    )
    error_message: Optional[str] = Field(
        default=None, description="Error message if failed"
    )

    model_config = ConfigDict(from_attributes=True)

    def is_successful(self) -> bool:
        """Check if activity completed successfully."""
        return self.status == "Succeeded"

    def is_failed(self) -> bool:
        """Check if activity failed."""
        return self.status == "Failed"



# ABOUTME: Notebook management MCP tools for Microsoft Fabric.
# ABOUTME: Provides tools for creating, updating, and executing notebooks.
"""Notebook management MCP tools.

This module provides MCP tools for Microsoft Fabric notebook operations including
creating notebooks and retrieving notebook definitions.
"""

from typing import Optional, TYPE_CHECKING
import logging

if TYPE_CHECKING:
    from fastmcp import FastMCP

from ..services import FabricNotebookService
from .base import handle_tool_errors, format_success_response, format_error_response, log_tool_invocation

logger = logging.getLogger(__name__)


def register_notebook_tools(mcp: "FastMCP", notebook_service: FabricNotebookService):
    """Register notebook management MCP tools.
    
    This function registers notebook-related tools for creating, updating, and
    inspecting notebooks, plus run history and driver logs.
    
    Args:
        mcp: FastMCP server instance to register tools on.
        notebook_service: Initialized FabricNotebookService instance.
        
    Example:
        ```python
        from ms_fabric_mcp_server import (
            FabricConfig, FabricClient,
            FabricWorkspaceService, FabricItemService, FabricNotebookService
        )
        from ms_fabric_mcp_server.tools import register_notebook_tools
        
        config = FabricConfig.from_environment()
        client = FabricClient(config)
        workspace_service = FabricWorkspaceService(client)
        item_service = FabricItemService(client)
        notebook_service = FabricNotebookService(client, item_service, workspace_service)
        
        register_notebook_tools(mcp, notebook_service)
        ```
    """
    
    @mcp.tool(title="Create Notebook")
    @handle_tool_errors
    def create_notebook(
        workspace_name: str,
        notebook_name: str,
        notebook_content: dict,
        description: Optional[str] = None,
        folder_path: Optional[str] = None,
        default_lakehouse_name: Optional[str] = None,
        lakehouse_workspace_name: Optional[str] = None,
    ) -> dict:
        """Create a notebook in a Fabric workspace.
        
        Parameters:
            workspace_name: The display name of the target workspace.
            notebook_name: Desired name inside Fabric (no folder separators).
            notebook_content: Notebook definition in ipynb JSON format.
            description: Optional description for the notebook.
            folder_path: Optional folder path (e.g., "demos/etl") to place the notebook.
                         Defaults to the workspace root when omitted.
            default_lakehouse_name: Optional default lakehouse to attach.
            lakehouse_workspace_name: Optional workspace name for the lakehouse.
            
        Returns:
            Dictionary with status, message, and notebook_id if successful.
        """
        log_tool_invocation(
            "create_notebook",
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            description=description,
            folder_path=folder_path,
            default_lakehouse_name=default_lakehouse_name,
            lakehouse_workspace_name=lakehouse_workspace_name,
        )
        logger.info(
            f"Creating notebook '{notebook_name}' in workspace '{workspace_name}'"
        )
        
        result = notebook_service.create_notebook(
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            notebook_content=notebook_content,
            description=description,
            folder_path=folder_path,
            default_lakehouse_name=default_lakehouse_name,
            lakehouse_workspace_name=lakehouse_workspace_name,
        )
        
        if result.status == "success":
            logger.info(f"Notebook created successfully: {result.notebook_id}")
            return {
                "status": "success",
                "message": result.message,
                "notebook_id": result.notebook_id,
            }
        logger.error(f"Notebook create failed: {result.message}")
        return {
            "status": "error",
            "message": result.message,
        }

    @mcp.tool(title="Get Notebook Definition")
    @handle_tool_errors
    def get_notebook_definition(
        workspace_name: str,
        notebook_name: str,
    ) -> dict:
        """Get the notebook definition (ipynb content)."""
        log_tool_invocation(
            "get_notebook_definition",
            workspace_name=workspace_name,
            notebook_name=notebook_name,
        )
        logger.info(
            f"Getting definition for notebook '{notebook_name}' in workspace '{workspace_name}'"
        )
        
        content = notebook_service.get_notebook_definition(workspace_name, notebook_name)
        
        result = {
            "status": "success",
            "workspace_name": workspace_name,
            "notebook_name": notebook_name,
            "definition": content,
        }
        
        logger.info("Successfully retrieved notebook definition")
        return result

    @mcp.tool(title="Update Notebook Definition")
    @handle_tool_errors
    def update_notebook_definition(
        workspace_name: str,
        notebook_name: str,
        notebook_content: Optional[dict] = None,
        default_lakehouse_name: Optional[str] = None,
        lakehouse_workspace_name: Optional[str] = None,
    ) -> dict:
        """Update notebook definition in Fabric.

        If notebook_content is omitted, the existing notebook definition is loaded
        and only metadata changes (e.g., default lakehouse) are applied.
        """
        log_tool_invocation(
            "update_notebook_definition",
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            default_lakehouse_name=default_lakehouse_name,
            lakehouse_workspace_name=lakehouse_workspace_name,
        )
        logger.info(
            f"Updating notebook '{notebook_name}' in workspace '{workspace_name}'"
        )

        result = notebook_service.update_notebook_definition(
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            notebook_content=notebook_content,
            default_lakehouse_name=default_lakehouse_name,
            lakehouse_workspace_name=lakehouse_workspace_name,
        )

        if result.status == "success":
            logger.info(f"Notebook updated successfully: {notebook_name}")
            return {
                "status": "success",
                "message": result.message,
                "notebook_id": result.notebook_id,
                "notebook_name": result.notebook_name,
                "workspace_id": result.workspace_id,
            }
        logger.error(f"Notebook update failed: {result.message}")
        return {
            "status": "error",
            "message": result.message,
        }
    
    @mcp.tool(title="Get Notebook Run Details")
    @handle_tool_errors
    def get_notebook_run_details(
        workspace_name: str,
        notebook_name: str,
        job_instance_id: str
    ) -> dict:
        """Get detailed run information for a notebook job instance.
        
        Retrieves execution metadata from the Fabric Notebook Livy Sessions API,
        which provides detailed timing, resource usage, and execution state information.
        
        **Use this tool when:**
        - You want to check the status and timing of a completed notebook run
        - You need to verify resource allocation for a notebook execution
        - You want to analyze execution performance (queue time, run time)
        
        **Note:** This method returns execution metadata (timing, state, resource usage).
        Cell-level outputs are only available for active sessions. Once a notebook job
        completes, individual cell outputs cannot be retrieved via the REST API. To
        capture cell outputs, use `mssparkutils.notebook.exit()` in your notebook and
        access the exitValue through Data Pipeline activities.
        
        Parameters:
            workspace_name: The display name of the workspace containing the notebook.
            notebook_name: Name of the notebook.
            job_instance_id: The job instance ID from execute_notebook or run_on_demand_job result.
            
        Returns:
            Dictionary with:
            - status: "success" or "error"
            - message: Description of the result
            - session: Full Livy session details (state, timing, resources)
            - execution_summary: Summarized execution information including:
                - state: Execution state (Success, Failed, Cancelled, etc.)
                - spark_application_id: Spark application identifier
                - queued_duration_seconds: Time spent in queue
                - running_duration_seconds: Actual execution time
                - total_duration_seconds: Total end-to-end time
                - driver_memory, driver_cores, executor_memory, etc.
            
        Example:
            ```python
            # After executing a notebook
            exec_result = run_on_demand_job(
                workspace_name="Analytics",
                item_name="ETL_Pipeline",
                item_type="Notebook",
                job_type="RunNotebook"
            )
            
            # Get detailed execution information
            details = get_notebook_run_details(
                workspace_name="Analytics",
                notebook_name="ETL_Pipeline",
                job_instance_id=exec_result["job_instance_id"]
            )
            
            if details["status"] == "success":
                summary = details["execution_summary"]
                print(f"State: {summary['state']}")
                print(f"Duration: {summary['total_duration_seconds']}s")
                print(f"Spark App ID: {summary['spark_application_id']}")
            ```
        """
        log_tool_invocation(
            "get_notebook_run_details",
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            job_instance_id=job_instance_id
        )
        logger.info(
            f"Getting run details for notebook '{notebook_name}' "
            f"job instance '{job_instance_id}'"
        )
        
        result = notebook_service.get_notebook_run_details(
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            job_instance_id=job_instance_id
        )
        
        if result.get("status") == "success":
            logger.info(f"Successfully retrieved run details for job instance '{job_instance_id}'")
        else:
            logger.error(f"Failed to get run details: {result.get('message')}")
        
        return result
    
    @mcp.tool(title="List Notebook Runs")
    @handle_tool_errors
    def list_notebook_runs(
        workspace_name: str,
        notebook_name: str,
        limit: Optional[int] = None
    ) -> dict:
        """List all Livy sessions (run history) for a notebook.
        
        Retrieves a list of all Livy sessions associated with a notebook, providing
        an execution history with job instance IDs, states, and timing information.
        
        **Use this tool when:**
        - You want to see the execution history of a notebook
        - You need to find a job instance ID for a past execution
        - You want to analyze execution patterns over time
        
        Parameters:
            workspace_name: The display name of the workspace containing the notebook.
            notebook_name: Name of the notebook.
            limit: Optional maximum number of sessions to return.
            
        Returns:
            Dictionary with:
            - status: "success" or "error"
            - message: Description of the result
            - sessions: List of session summaries, each containing:
                - job_instance_id: Unique identifier for the job
                - livy_id: Livy session identifier
                - state: Execution state (Success, Failed, Cancelled, etc.)
                - operation_name: Type of operation (Notebook Scheduled Run, etc.)
                - spark_application_id: Spark application identifier
                - submitted_time_utc: When the job was submitted
                - start_time_utc: When execution started
                - end_time_utc: When execution ended
                - total_duration_seconds: Total execution time
            - total_count: Total number of sessions found
            
        Example:
            ```python
            history = list_notebook_runs(
                workspace_name="Analytics",
                notebook_name="ETL_Pipeline",
                limit=10
            )
            
            if history["status"] == "success":
                print(f"Found {history['total_count']} executions")
                for session in history["sessions"]:
                    print(f"{session['job_instance_id']}: {session['state']}")
            ```
        """
        log_tool_invocation(
            "list_notebook_runs",
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            limit=limit
        )
        logger.info(
            f"Listing runs for notebook '{notebook_name}' "
            f"in workspace '{workspace_name}'"
        )
        
        result = notebook_service.list_notebook_runs(
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            limit=limit
        )
        
        if result.get("status") == "success":
            logger.info(f"Found {result.get('total_count', 0)} runs for notebook '{notebook_name}'")
        else:
            logger.error(f"Failed to list runs: {result.get('message')}")
        
        return result
    
    @mcp.tool(title="Get Notebook Driver Logs")
    @handle_tool_errors
    def get_notebook_driver_logs(
        workspace_name: str,
        notebook_name: str,
        job_instance_id: str,
        log_type: str = "stdout",
        max_lines: Optional[int] = 500
    ) -> dict:
        """Get Spark driver logs for a notebook execution.
        
        Retrieves the driver logs (stdout or stderr) from a completed notebook run.
        This is particularly useful for getting detailed error messages and Python
        tracebacks when a notebook fails.
        
        **Important Notes**:
        - Python exceptions and tracebacks appear in `stdout`, not `stderr`
        - `stderr` contains Spark/system logs (typically larger)
        - For failed notebooks, check `stdout` first for the Python error
        - Look for "Error", "Exception", "Traceback" in the output
        
        **Use this tool when:**
        - A notebook execution failed and you need to see the Python error
        - You want to debug notebook issues by examining driver logs
        - You need to analyze Spark driver behavior (stderr)
        
        Parameters:
            workspace_name: The display name of the workspace containing the notebook.
            notebook_name: Name of the notebook.
            job_instance_id: The job instance ID from execute_notebook or run_on_demand_job result.
            log_type: Type of log to retrieve - "stdout" (default) or "stderr".
                     Use "stdout" for Python errors and print statements.
                     Use "stderr" for Spark/system logs.
            max_lines: Maximum number of lines to return (default: 500, None for all).
                      Returns the last N lines (most recent, where errors typically are).
            
        Returns:
            Dictionary with:
            - status: "success" or "error"
            - message: Description of the result
            - log_type: Type of log retrieved
            - log_content: The actual log content as a string
            - log_size_bytes: Total size of the log file
            - truncated: Whether the log was truncated
            - spark_application_id: The Spark application ID
            - livy_id: The Livy session ID
            
        Example:
            ```python
            # Get Python error from a failed notebook
            result = get_notebook_driver_logs(
                workspace_name="Analytics",
                notebook_name="ETL_Pipeline",
                job_instance_id="12345678-1234-1234-1234-123456789abc",
                log_type="stdout"  # Python errors are in stdout!
            )
            
            if result["status"] == "success":
                print(result["log_content"])
                # Output will include Python traceback like:
                # ZeroDivisionError: division by zero
                # Traceback (most recent call last):
                #   Cell In[11], line 2
                #     result = x / 0
            ```
        """
        log_tool_invocation(
            "get_notebook_driver_logs",
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            job_instance_id=job_instance_id,
            log_type=log_type,
            max_lines=max_lines
        )
        logger.info(
            f"Getting driver logs ({log_type}) for notebook '{notebook_name}' "
            f"job instance '{job_instance_id}'"
        )
        
        result = notebook_service.get_notebook_driver_logs(
            workspace_name=workspace_name,
            notebook_name=notebook_name,
            job_instance_id=job_instance_id,
            log_type=log_type,
            max_lines=max_lines
        )
        
        if result.get("status") == "success":
            logger.info(
                f"Successfully retrieved {log_type} logs "
                f"({result.get('log_size_bytes', 0)} bytes) "
                f"for job instance '{job_instance_id}'"
            )
        else:
            logger.error(f"Failed to get driver logs: {result.get('message')}")
        
        return result
    
    logger.info("Notebook tools registered successfully (6 tools)")
